import pytest

from ...caluma_core.tests import extract_serializer_input_fields
from .. import models, serializers


@pytest.mark.parametrize("task__type", [models.Task.TYPE_SIMPLE])
def test_query_all_tasks(db, snapshot, task, schema_executor):
    query = """
        query AllTasks($name: String!) {
          allTasks(name: $name) {
            totalCount
            edges {
              node {
                __typename
                slug
                name
                description
                meta
              }
            }
          }
        }
    """

    result = schema_executor(query, variables={"name": task.name})

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("mutation", ["SaveSimpleTask", "SaveCompleteWorkflowFormTask"])
def test_save_task(db, snapshot, task, mutation, schema_executor):
    mutation_func = mutation[0].lower() + mutation[1:]
    query = f"""
        mutation {mutation}($input: {mutation}Input!) {{
          {mutation_func}(input: $input) {{
            task {{
                slug
                name
                __typename
                meta
            }}
            clientMutationId
          }}
        }}
    """

    inp = {
        "input": extract_serializer_input_fields(serializers.SaveTaskSerializer, task)
    }
    result = schema_executor(query, variables=inp)
    assert not result.errors
    snapshot.assert_match(result.data)


def test_save_comlete_task_form_task(db, snapshot, task, schema_executor):
    query = """
        mutation SaveCompleteTaskFormTask($input: SaveCompleteTaskFormTaskInput!) {
          saveCompleteTaskFormTask(input: $input) {
            task {
                slug
                name
                __typename
                meta
            }
            clientMutationId
          }
        }
    """

    inp = {
        "input": extract_serializer_input_fields(
            serializers.SaveCompleteTaskFormTaskSerializer, task
        )
    }
    result = schema_executor(query, variables=inp)
    assert not result.errors
    snapshot.assert_match(result.data)



from django.conf import settings
from django.core.management.base import BaseCommand
from minio.error import BucketAlreadyExists, BucketAlreadyOwnedByYou, ResponseError

from ...storage_clients import client


class Command(BaseCommand):
    """Create a bucket based on the config."""

    help = "Create a bucket based on the config."

    def handle(self, *args, **options):
        bucket_name = settings.MINIO_STORAGE_MEDIA_BUCKET_NAME

        try:
            client.client.make_bucket(bucket_name)
        except BucketAlreadyOwnedByYou:  # pragma: no cover
            self.stderr.write(f"Bucket: {bucket_name} is already owned by you!")
        except BucketAlreadyExists:  # pragma: no cover
            self.stderr.write(f"Bucket: {bucket_name} already exists!")
        except ResponseError:  # pragma: no cover
            raise
        else:
            self.stdout.write(f"Created bucket: {bucket_name}")



import os

import pytest
from django.conf import settings
from django.core.management import call_command
from django.utils import timezone
from minio import Minio
from simple_history.models import registered_models

from caluma.caluma_core.management.commands import cleanup_history

from ..models import Form


def test_create_bucket_command(mocker):
    mocker.patch.object(Minio, "make_bucket")
    call_command("create_bucket", stdout=open(os.devnull, "w"))
    Minio.make_bucket.assert_called_once_with(settings.MINIO_STORAGE_MEDIA_BUCKET_NAME)


@pytest.mark.parametrize("force", [True, False])
@pytest.mark.parametrize("keep,kept", [("1 year", 2), ("1 day", 1), (None, 2)])
def test_cleanup_history_command(db, force, keep, kept):
    # we need to override the registered models dict in order to get rid of the
    # fake models created in caluma_core tests
    cleanup_history.registered_models = {
        k: v for k, v in registered_models.items() if not k.startswith("caluma_core_")
    }
    kwargs = {}
    if force:
        kwargs["force"] = force
    else:
        kept = 3
    if keep:
        kwargs["keep"] = keep

    Form.objects.create(slug="form 1")

    f2 = Form.objects.create(slug="form 2")
    f2_hist = f2.history.first()
    f2_hist.history_date = f2_hist.history_date - timezone.timedelta(days=2)
    f2_hist.save()

    f3 = Form.objects.create(slug="form 3")
    f3_hist = f3.history.first()
    f3_hist.history_date = f3_hist.history_date - timezone.timedelta(days=730)
    f3_hist.save()

    call_command("cleanup_history", **kwargs, stdout=open(os.devnull, "w"))

    assert Form.history.count() == kept



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_complete_work_item_last[ready-completed-True-simple-None] 1"] = {
    "completeWorkItem": {
        "clientMutationId": None,
        "workItem": {
            "case": {"closedByUser": "admin", "status": "COMPLETED"},
            "closedByUser": "admin",
            "status": "COMPLETED",
        },
    }
}

snapshots["test_complete_multiple_instance_task_form_work_item_next[integer-1] 1"] = {
    "completeWorkItem": {
        "clientMutationId": None,
        "workItem": {
            "case": {
                "status": "RUNNING",
                "workItems": {
                    "edges": [
                        {
                            "node": {
                                "addressedGroups": ["group-name"],
                                "status": "READY",
                            }
                        },
                        {"node": {"addressedGroups": [], "status": "COMPLETED"}},
                        {"node": {"addressedGroups": [], "status": "COMPLETED"}},
                    ],
                    "totalCount": 3,
                },
            },
            "status": "COMPLETED",
        },
    }
}

snapshots["test_complete_work_item_with_next[ready-None-simple] 1"] = {
    "completeWorkItem": {
        "clientMutationId": None,
        "workItem": {
            "case": {
                "status": "RUNNING",
                "workItems": {
                    "edges": [
                        {
                            "node": {
                                "addressedGroups": ["group-name"],
                                "status": "READY",
                            }
                        },
                        {"node": {"addressedGroups": [], "status": "COMPLETED"}},
                    ],
                    "totalCount": 2,
                },
            },
            "status": "COMPLETED",
        },
    }
}



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots[
    "test_base_format_validators[question__format_validators1-test@example.com-True-text] 1"
] = {
    "saveDocumentStringAnswer": {
        "answer": {"stringValue": "test@example.com"},
        "clientMutationId": "testid",
    }
}

snapshots[
    "test_base_format_validators[question__format_validators1-test@example.com-True-textarea] 1"
] = {
    "saveDocumentStringAnswer": {
        "answer": {"stringValue": "test@example.com"},
        "clientMutationId": "testid",
    }
}

snapshots[
    "test_base_format_validators[question__format_validators3-+411234567890-True-text] 1"
] = {
    "saveDocumentStringAnswer": {
        "answer": {"stringValue": "+411234567890"},
        "clientMutationId": "testid",
    }
}

snapshots[
    "test_base_format_validators[question__format_validators3-+411234567890-True-textarea] 1"
] = {
    "saveDocumentStringAnswer": {
        "answer": {"stringValue": "+411234567890"},
        "clientMutationId": "testid",
    }
}

snapshots["test_fetch_format_validators 1"] = {
    "allFormatValidators": {
        "edges": [
            {
                "node": {
                    "errorMsg": "Not valid",
                    "name": "test name english",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "test-validator",
                }
            },
            {
                "node": {
                    "errorMsg": "Please enter a valid Email address.",
                    "name": "E-mail",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "email",
                }
            },
            {
                "node": {
                    "errorMsg": "Please enter a valid phone number.",
                    "name": "Phone number",
                    "regex": "^[\\s\\/\\.\\(\\)-]*(?:\\+|0|00)(?:[\\s\\/\\.\\(\\)-]*\\d[\\s\\/\\.\\(\\)-]*){6,20}$",
                    "slug": "phone-number",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_format_validators 2"] = {
    "allFormatValidators": {
        "edges": [
            {
                "node": {
                    "errorMsg": "Nicht valid",
                    "name": "test name deutsch",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "test-validator",
                }
            },
            {
                "node": {
                    "errorMsg": "Bitte geben Sie eine gültige E-Mail-Adresse ein.",
                    "name": "E-Mail",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "email",
                }
            },
            {
                "node": {
                    "errorMsg": "Bitte geben Sie eine gültige Telefonnummer ein.",
                    "name": "Telefonnummer",
                    "regex": "^[\\s\\/\\.\\(\\)-]*(?:\\+|0|00)(?:[\\s\\/\\.\\(\\)-]*\\d[\\s\\/\\.\\(\\)-]*){6,20}$",
                    "slug": "phone-number",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_format_validators 3"] = {
    "allFormatValidators": {
        "edges": [
            {
                "node": {
                    "errorMsg": "Not valid",
                    "name": "test name english",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "test-validator",
                }
            },
            {
                "node": {
                    "errorMsg": "Veuillez entrer une addresse e-mail valide.",
                    "name": "Courriel",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "email",
                }
            },
            {
                "node": {
                    "errorMsg": "Veuillez entrer un numéro de téléphone valide.",
                    "name": "numéro de téléphone",
                    "regex": "^[\\s\\/\\.\\(\\)-]*(?:\\+|0|00)(?:[\\s\\/\\.\\(\\)-]*\\d[\\s\\/\\.\\(\\)-]*){6,20}$",
                    "slug": "phone-number",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}

snapshots["test_fetch_format_validators 4"] = {
    "allFormatValidators": {
        "edges": [
            {
                "node": {
                    "errorMsg": "Not valid",
                    "name": "test name english",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "test-validator",
                }
            },
            {
                "node": {
                    "errorMsg": "Please enter a valid Email address.",
                    "name": "E-mail",
                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                    "slug": "email",
                }
            },
            {
                "node": {
                    "errorMsg": "Please enter a valid phone number.",
                    "name": "Phone number",
                    "regex": "^[\\s\\/\\.\\(\\)-]*(?:\\+|0|00)(?:[\\s\\/\\.\\(\\)-]*\\d[\\s\\/\\.\\(\\)-]*){6,20}$",
                    "slug": "phone-number",
                }
            },
        ],
        "pageInfo": {
            "endCursor": "YXJyYXljb25uZWN0aW9uOjI=",
            "startCursor": "YXJyYXljb25uZWN0aW9uOjA=",
        },
        "totalCount": 3,
    }
}



from functools import reduce

import graphene
from django import forms
from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.contrib.postgres.fields.hstore import KeyTransform
from django.contrib.postgres.fields.jsonb import KeyTextTransform
from django.contrib.postgres.search import SearchVector
from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.db.models.expressions import OrderBy, RawSQL
from django.db.models.functions import Cast
from django.utils import translation
from django_filters.constants import EMPTY_VALUES
from django_filters.fields import ChoiceField
from django_filters.rest_framework import (
    CharFilter,
    Filter,
    FilterSet,
    MultipleChoiceFilter,
    OrderingFilter,
)
from graphene import Enum, InputObjectType, List
from graphene.types import generic
from graphene.types.utils import get_type
from graphene.utils.str_converters import to_camel_case
from graphene_django import filter
from graphene_django.converter import convert_choice_name
from graphene_django.filter.filterset import GrapheneFilterSetMixin
from graphene_django.forms.converter import convert_form_field
from graphene_django.registry import get_global_registry
from localized_fields.fields import LocalizedField

from .forms import (
    GlobalIDFormField,
    GlobalIDMultipleChoiceField,
    SlugMultipleChoiceField,
)
from .ordering import CalumaOrdering
from .relay import extract_global_id
from .types import DjangoConnectionField


class CompositeFieldClass(forms.MultiValueField):
    """Mixin to build complex field classes.

    This is just to pretend to Graphene that it's a composite type.
    It's the base of the internal representation that only passes
    values from the request down to the filters (or similar).

    The actual schema type is generated via the `convert_form_field()`
    function from `graphene_django.forms.converter`.
    """

    def __init__(self, label, *, fields=None, **kwargs):
        fields = (forms.CharField(), forms.CharField())
        super().__init__(fields=fields)

    def clean(self, data):
        # override parent clean() which would reject our data structure.
        # We don't validate, as the structure is already enforced by the
        # schema.

        return data


class AscDesc(Enum):
    ASC = "ASC"
    DESC = "DESC"


class FilterCollectionFilter(Filter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        filter_coll = self.filterset_class()
        for flt in value:
            invert = flt.pop("invert", False)
            flt_key = list(flt.keys())[0]
            flt_val = flt[flt_key]
            filter = filter_coll.get_filters()[flt_key]

            new_qs = filter.filter(qs, flt_val)

            if invert:
                qs = qs.exclude(pk__in=new_qs)
            else:
                qs = new_qs

        return qs


class FilterCollectionOrdering(Filter):
    def _order_part(self, qs, ord, filter_coll):
        direction = ord.pop("direction", "ASC")

        assert len(ord) == 1
        filt_name = list(ord.keys())[0]
        filter = filter_coll.get_filters()[filt_name]
        qs, field = filter.get_ordering_value(qs, ord[filt_name])

        # Normally, people use ascending order, and in this context it seems
        # natural to have NULL entries at the end.
        # Making the `nulls_first`/`nulls_last` parameter accessible in the
        # GraphQL interface would be overkill, at least for now.
        return (
            qs,
            OrderBy(
                field,
                descending=(direction == "DESC"),
                nulls_first=(direction == "DESC"),
                nulls_last=(direction == "ASC"),
            ),
        )

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs
        filter_coll = self.filterset_class()

        order_by = []
        for ord in value:
            qs, order_field = self._order_part(qs, ord, filter_coll)
            order_by.append(order_field)

        if order_by:
            qs = qs.order_by(*order_by)

        return qs


def FilterCollectionFactory(filterset_class, ordering):  # noqa:C901
    """
    Build a single filter from a `FilterSet`.

    This converts an arbitrary `FilterSet` class into a single filter that
    allows chaining of filters as a list. In addition, this introduces
    an optional `invert` parameter, allowing requestors to use
    a filter for either inclusion or exclusion.

    On the schema side, this generates a new type to represent the filter value
    whose name is derived from the given filterset class.

    Usage:
    >>> class MyFilterSet(FilterSet):
    ...     filter = FilterCollectionFactory(FilterSetWithActualFilters, ordering=False)

    """

    field_class_name = f"{filterset_class.__name__}Field"
    field_type_name = f"{filterset_class.__name__}Type"
    collection_name = f"{filterset_class.__name__}Collection"

    # The field class.
    custom_field_class = type(field_class_name, (CompositeFieldClass,), {})

    @convert_form_field.register(custom_field_class)
    def convert_field(field):

        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            return converted

        _filter_coll = filterset_class()

        def _get_or_make_field(name, filt):
            return convert_form_field(filt.field)

        def _should_include_filter(filt):
            # if we're in ordering mode, we want
            # to return True for all CalumaOrdering types,
            # and if it's false, we want the opposite
            return ordering == isinstance(filt, CalumaOrdering)

        filter_fields = {
            name: _get_or_make_field(name, filt)
            for name, filt in _filter_coll.filters.items()
            # exclude orderBy in our fields. We want only new-style order filters
            if _should_include_filter(filt) and name != "orderBy"
        }

        if ordering:
            filter_fields["direction"] = AscDesc(default=AscDesc.ASC, required=False)
        else:
            filter_fields["invert"] = graphene.Boolean(required=False, default=False)

        filter_type = type(field_type_name, (InputObjectType,), filter_fields)

        converted = List(filter_type)
        registry.register_converted_field(field, converted)
        return converted

    filter_impl = FilterCollectionOrdering if ordering else FilterCollectionFilter

    filter_coll = type(
        collection_name,
        (filter_impl,),
        {"field_class": custom_field_class, "filterset_class": filterset_class},
    )
    return filter_coll()


def CollectionFilterSetFactory(filterset_class, orderset_class=None):
    """
    Build single-filter filterset classes.

    Use this in place of a regular filterset_class parametrisation in
    the serializers.
    If you pass the optional `orderset_class` parameter, it is used for
    an `order` filter. The filters defined in the `orderset_class` must
    inherit from `caluma.caluma_core.ordering.CalumaOrdering` and provide a
    `get_ordering_value()` method.

    Example:
    >>> all_documents = DjangoFilterConnectionField(
    ...    Document, filterset_class=CollectionFilterSetFactory(DocumentFilterSet)
    ... )

    """

    suffix = "" if orderset_class else "NoOrdering"
    cache_key = filterset_class.__name__ + suffix

    if cache_key in CollectionFilterSetFactory._cache:
        return CollectionFilterSetFactory._cache[cache_key]

    coll_fields = {"filter": FilterCollectionFactory(filterset_class, ordering=False)}
    if orderset_class:
        coll_fields["order"] = FilterCollectionFactory(orderset_class, ordering=True)

    ret = CollectionFilterSetFactory._cache[cache_key] = type(
        f"{filterset_class.__name__}Collection",
        (filterset_class, FilterSet),
        {
            **coll_fields,
            "Meta": type(
                "Meta",
                (filterset_class.Meta,),
                {
                    "model": filterset_class.Meta.model,
                    "fields": filterset_class.Meta.fields + tuple(coll_fields.keys()),
                },
            ),
        },
    )

    return ret


CollectionFilterSetFactory._cache = {}


class GlobalIDFilter(Filter):
    field_class = GlobalIDFormField

    def filter(self, qs, value):
        _id = None
        if value is not None:
            _id = extract_global_id(value)
        return super(GlobalIDFilter, self).filter(qs, _id)


class GlobalIDMultipleChoiceFilter(MultipleChoiceFilter):
    field_class = GlobalIDMultipleChoiceField

    def filter(self, qs, value):
        gids = [extract_global_id(v) for v in value]
        return super(GlobalIDMultipleChoiceFilter, self).filter(qs, gids)


class SlugMultipleChoiceFilter(MultipleChoiceFilter):
    field_class = SlugMultipleChoiceField

    def filter(self, qs, value):
        return super().filter(qs, value)


class LocalizedFilter(Filter):
    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        lang = translation.get_language()
        filter_expr = "{0}__{1}__{2}".format(self.field_name, lang, self.lookup_expr)
        return qs.filter(**{filter_expr: value})


GrapheneFilterSetMixin.FILTER_DEFAULTS.update(
    {
        LocalizedField: {"filter_class": LocalizedFilter},
        models.AutoField: {"filter_class": GlobalIDFilter},
        models.OneToOneField: {"filter_class": GlobalIDFilter},
        models.ForeignKey: {"filter_class": GlobalIDFilter},
        models.ManyToManyField: {"filter_class": GlobalIDMultipleChoiceFilter},
        models.ManyToOneRel: {"filter_class": GlobalIDMultipleChoiceFilter},
        models.ManyToManyRel: {"filter_class": GlobalIDMultipleChoiceFilter},
    }
)


class SearchFilter(Filter):
    """
    Enable fulltext search on queryset.

    Define fields which need to be searched in.
    """

    def __init__(self, *args, fields, **kwargs):
        self.fields = fields
        super().__init__(*args, **kwargs)

    def _get_model_field(self, model, field):
        model_field = model._meta.get_field(field)
        return model_field, getattr(model_field, "related_model", None)

    def _build_search_expression(self, field_lookup):
        # TODO: is there no Django API which allows conversion of lookup to django field?
        model_field, _ = reduce(
            lambda model_tuple, field: self._get_model_field(model_tuple[1], field),
            field_lookup.split(LOOKUP_SEP),
            (None, self.model),
        )

        if isinstance(model_field, LocalizedField):
            lang = translation.get_language()
            return KeyTransform(lang, field_lookup)
        elif isinstance(model_field, JSONField):
            return Cast(field_lookup, models.TextField())

        return field_lookup

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        qs = qs.annotate(
            search=SearchVector(
                *[self._build_search_expression(field) for field in self.fields]
            )
        )

        return qs.filter(search=value)


class OrderingField(ChoiceField):
    """
    Specific ordering field used as marker to convert to graphql type.

    See `convert_ordering_field_to_enum`
    """

    def validate(self, value):
        invalid = set(value or []) - {choice[0] for choice in self.choices}
        return not bool(invalid)

    def to_python(self, value):
        return value


class ListField(forms.Field):
    """List field as to allow actual lists in ordering vs csv string."""

    pass


class OrderingFilter(OrderingFilter):
    """Ordering filter adding default fields from models.BaseModel.

    Label is required and is used for enum naming in GraphQL schema.

    This filter additionally allows sorting by meta field values.
    """

    base_field_class = ListField
    field_class = OrderingField

    def __init__(self, label, *args, fields=tuple(), **kwargs):
        fields = tuple(fields) + (
            "created_at",
            "modified_at",
            "created_by_user",
            "created_by_group",
            *[f"meta_{f}" for f in settings.META_FIELDS],
        )

        super().__init__(
            *args,
            fields=fields,
            label=label,
            empty_label=None,
            null_label=None,
            **kwargs,
        )

    def get_ordering_value(self, param):
        if not any(param.startswith(prefix) for prefix in ("meta_", "-meta_")):
            return super().get_ordering_value(param)

        descending = False
        if param.startswith("-"):
            descending = True
            param = param[1:]

        meta_field_key = param[5:]

        # order_by works on json field keys only without dashes
        # but we want to support dasherized keys as well as this is
        # valid json, hence need to use raw sql
        return OrderBy(
            RawSQL(f'"{self.model._meta.db_table}"."meta"->%s', (meta_field_key,)),
            descending=descending,
        )


class IntegerFilter(Filter):
    field_class = forms.IntegerField


class FilterSet(GrapheneFilterSetMixin, FilterSet):
    created_by_user = CharFilter()
    created_by_group = CharFilter()


class JSONLookupMode(Enum):
    EXACT = "exact"
    STARTSWITH = "startswith"
    CONTAINS = "contains"
    ICONTAINS = "icontains"
    GTE = "gte"
    GT = "gt"
    LTE = "lte"
    LT = "lt"


class JSONValueFilterType(InputObjectType):
    key = graphene.String(required=True)
    value = generic.GenericScalar(required=True)
    lookup = JSONLookupMode()


class JSONValueFilterField(CompositeFieldClass):
    pass


class JSONValueFilter(Filter):
    field_class = JSONValueFilterField

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        for expr in value:
            if expr in EMPTY_VALUES:  # pragma: no cover
                continue

            lookup_expr = expr.get("lookup", self.lookup_expr)
            # "contains" behaves differently on JSONFields as it does on TextFields.
            # That's why we annotate the queryset with the value.
            # Some discussion about it can be found here:
            # https://code.djangoproject.com/ticket/26511
            if isinstance(expr["value"], str):
                qs = qs.annotate(
                    field_val=KeyTextTransform(expr["key"], self.field_name)
                )
                lookup = {f"field_val__{lookup_expr}": expr["value"]}
            else:
                lookup = {
                    f"{self.field_name}__{expr['key']}__{lookup_expr}": expr["value"]
                }
            qs = qs.filter(**lookup)
        return qs

    @staticmethod
    @convert_form_field.register(JSONValueFilterField)
    def convert_meta_value_field(field):
        registry = get_global_registry()
        converted = registry.get_converted_field(field)
        if converted:
            return converted

        converted = List(JSONValueFilterType)
        registry.register_converted_field(field, converted)
        return converted


class MetaFilterSet(FilterSet):
    meta_has_key = CharFilter(lookup_expr="has_key", field_name="meta")
    meta_value = JSONValueFilter(field_name="meta")


class DjangoFilterConnectionField(
    filter.DjangoFilterConnectionField, DjangoConnectionField
):
    @property
    def filterset_class(self):
        return self._provided_filterset_class


class DjangoFilterSetConnectionField(DjangoFilterConnectionField):
    @property
    def model(self):
        return self.filterset_class._meta.model

    @property
    def type(self):
        return get_type(self._type)


@convert_form_field.register(OrderingField)
def convert_ordering_field_to_enum(field):
    """
    Add support to convert ordering choices to Graphql enum.

    Label is used as enum name.
    """
    registry = get_global_registry()
    converted = registry.get_converted_field(field)

    if converted:
        return converted

    if field.label in convert_ordering_field_to_enum._cache:
        return convert_ordering_field_to_enum._cache[field.label]

    def get_choices(choices):
        for value, help_text in choices:
            if value[0] != "-":
                name = convert_choice_name(value) + "_ASC"
            else:
                name = convert_choice_name(value[1:]) + "_DESC"
            description = help_text
            yield name, value, description

    name = to_camel_case(field.label)
    choices = list(get_choices(field.choices))
    named_choices = [(c[0], c[1]) for c in choices]
    named_choices_descriptions = {c[0]: c[2] for c in choices}

    class EnumWithDescriptionsType(object):
        @property
        def description(self):
            return named_choices_descriptions[self.name]

    enum = Enum(name, list(named_choices), type=EnumWithDescriptionsType)
    converted = List(enum, description=field.help_text, required=field.required)

    registry.register_converted_field(field, converted)
    convert_ordering_field_to_enum._cache[field.label] = converted
    return converted


convert_ordering_field_to_enum._cache = {}


@convert_form_field.register(forms.ChoiceField)
@convert_form_field.register(ChoiceField)
def convert_choice_field_to_enum(field):
    """
    Add support to convert ordering choices to Graphql enum.

    Label is used as enum name.
    """

    registry = get_global_registry()

    # field label of enum needs to be unique so stored it likewise
    converted = registry.get_converted_field(field.label)
    if converted:
        return converted

    def get_choices(choices):
        for value, help_text in choices:
            if value:
                name = convert_choice_name(value)
                description = help_text
                yield name, value, description

    name = to_camel_case(field.label)
    choices = list(get_choices(field.choices))
    named_choices = [(c[0], c[1]) for c in choices]
    named_choices_descriptions = {c[0]: c[2] for c in choices}

    class EnumWithDescriptionsType(object):
        @property
        def description(self):
            return named_choices_descriptions[self.name]

    enum = Enum(name, list(named_choices), type=EnumWithDescriptionsType)
    converted = enum(description=field.help_text, required=field.required)

    registry.register_converted_field(field.label, converted)
    return converted


def generate_list_filter_class(inner_type):
    """
    Return a Filter class that will resolve into a List(`inner_type`) graphene type.

    This allows us to do things like use `__in` and `__overlap` filters that accept
    graphene lists instead of a comma delimited value string that's interpolated into
    a list by django_filters.BaseCSVFilter (which is used to define
    django_filters.BaseInFilter)
    """

    form_field = type(f"List{inner_type.__name__}FormField", (forms.Field,), {})
    filter_class = type(
        f"{inner_type.__name__}ListFilter",
        (Filter,),
        {
            "field_class": form_field,
            "__doc__": (
                f"{inner_type.__name__}ListFilter is a small extension of a raw "
                f"django_filters.Filter that allows us to express graphql "
                f"List({inner_type.__name__}) arguments using FilterSets. "
                f"Note that the given values are passed directly into queryset filters."
            ),
        },
    )
    convert_form_field.register(form_field)(
        lambda x: graphene.List(inner_type, required=x.required)
    )

    return filter_class


StringListFilter = generate_list_filter_class(graphene.String)



import pytest
from minio import Minio

from ...caluma_form.models import File, Question


def test_delete_file_answer(
    db,
    question_factory,
    answer_factory,
    form_question_factory,
    file_factory,
    form,
    document,
    mocker,
):
    file_question = question_factory(type=Question.TYPE_FILE)
    form_question_factory(question=file_question, form=form)
    answer = answer_factory(
        question=file_question, value=None, document=document, file=file_factory()
    )
    mocker.patch.object(Minio, "copy_object")
    mocker.patch.object(Minio, "remove_object")
    answer.delete()
    with pytest.raises(File.DoesNotExist):
        answer.file.refresh_from_db()


def test_update_file(db, file_factory, mocker):
    file = file_factory()
    mocker.patch.object(Minio, "copy_object")
    mocker.patch.object(Minio, "remove_object")
    file.name = "something else"
    file.save()
    Minio.copy_object.assert_called()
    Minio.remove_object.assert_called()



import inspect
from functools import wraps

from django.core.exceptions import ImproperlyConfigured

from .collections import list_duplicates


def validation_for(mutation):
    """Decorate function to overwriting validation of specific mutation."""

    def decorate(fn):
        @wraps(fn)
        def _decorated(self, mutation, data, info):
            return fn(self, mutation, data, info)

        _decorated._validation = mutation
        return _decorated

    return decorate


class BaseValidation(object):
    """Basic validation class to be extended by any validation implementation.

    In combination with the decorators `@validation_for` a custom validation class
    can define validations for specific mutations.
    It is also possible to execute the same validation logic for all mutations
    using the base class `Mutation`.

    A custom validation class could look like this:
    ```
    >>> from caluma.caluma_form.schema import SaveForm
    ... from caluma.mutation import Mutation
    ... from rest_framework import exceptions
    ...
    ... class CustomValidation(BaseValidation):
    ...     @validation_for(Mutation)
    ...     def validate_mutation(self, mutation, data, info):
    ...         # add your default specific validation code here
    ...         return data
    ...
    ...     @validation_for(SaveForm)
    ...     def validate_save_form(self, mutation, data, info):
    ...         if data['meta'] and info.context.group != 'admin':
    ...           raise exceptions.ValidationException('May not change meta on form')
    ...         return data
    """

    def __init__(self):
        validation_fns = inspect.getmembers(self, lambda m: hasattr(m, "_validation"))
        validation_muts = [fn._validation.__name__ for _, fn in validation_fns]
        validation_muts_dups = list_duplicates(validation_muts)
        if validation_muts_dups:
            raise ImproperlyConfigured(
                f"`validation_for` defined multiple times for "
                f"{', '.join(validation_muts_dups)} in {str(self)}"
            )
        self._validations = {fn._validation: fn for _, fn in validation_fns}

    def validate(self, mutation, data, info):
        for cls in mutation.mro():
            if cls in self._validations:
                return self._validations[cls](mutation, data, info)
        return data



from functools import partial
from itertools import chain

from pyjexl.analysis import ValidatingAnalyzer

from ..caluma_core.jexl import JEXL, ExtractTransformSubjectAnalyzer


class GroupValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "groups" and not isinstance(transform.subject.value, list):
            yield f"{transform.subject.value} is not a valid list of groups."

        yield from super().visit_Transform(transform)


class TaskValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "tasks" and not isinstance(transform.subject.value, list):
            yield f"{transform.subject.value} is not a valid list of tasks."

        if transform.name == "task" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid task slug."

        yield from super().visit_Transform(transform)


class GroupJexl(JEXL):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_transform("groups", lambda spec: spec)

    def validate(self, expression):
        return super().validate(expression, GroupValidatingAnalyzer)


class FlowJexl(JEXL):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_transform("task", lambda spec: spec)
        self.add_transform("tasks", lambda spec: spec)

    def validate(self, expression):
        return super().validate(expression, TaskValidatingAnalyzer)

    def extract_tasks(self, expr):
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=["task"])
        )

        # tasks transforms return a list of literals
        yield from (
            literal.value
            for literal in chain(
                *self.analyze(
                    expr, partial(ExtractTransformSubjectAnalyzer, transforms=["tasks"])
                )
            )
        )



from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.postgres.indexes import GinIndex
from django.db import models, transaction
from django.db.models.signals import post_init
from django.dispatch import receiver
from localized_fields.fields import LocalizedField, LocalizedTextField

from ..caluma_core.models import NaturalKeyModel, SlugModel, UUIDModel
from .storage_clients import client


class Form(SlugModel):
    name = LocalizedField(blank=False, null=False, required=False)
    description = LocalizedField(blank=True, null=True, required=False)
    meta = JSONField(default=dict)
    is_published = models.BooleanField(default=False)
    is_archived = models.BooleanField(default=False)
    questions = models.ManyToManyField(
        "Question", through="FormQuestion", related_name="forms"
    )
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        help_text="Reference this form has been copied from",
        related_name="+",
        on_delete=models.SET_NULL,
    )

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class FormQuestion(NaturalKeyModel):
    form = models.ForeignKey("Form", on_delete=models.CASCADE)
    question = models.ForeignKey("Question", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    def natural_key(self):
        return f"{self.form_id}.{self.question_id}"

    class Meta:
        ordering = ["-sort"]
        unique_together = ("form", "question")


class Question(SlugModel):
    TYPE_MULTIPLE_CHOICE = "multiple_choice"
    TYPE_INTEGER = "integer"
    TYPE_FLOAT = "float"
    TYPE_DATE = "date"
    TYPE_CHOICE = "choice"
    TYPE_TEXTAREA = "textarea"
    TYPE_TEXT = "text"
    TYPE_TABLE = "table"
    TYPE_FORM = "form"
    TYPE_FILE = "file"
    TYPE_DYNAMIC_CHOICE = "dynamic_choice"
    TYPE_DYNAMIC_MULTIPLE_CHOICE = "dynamic_multiple_choice"
    TYPE_STATIC = "static"

    TYPE_CHOICES = (
        TYPE_MULTIPLE_CHOICE,
        TYPE_INTEGER,
        TYPE_FLOAT,
        TYPE_DATE,
        TYPE_CHOICE,
        TYPE_TEXTAREA,
        TYPE_TEXT,
        TYPE_TABLE,
        TYPE_FORM,
        TYPE_FILE,
        TYPE_DYNAMIC_CHOICE,
        TYPE_DYNAMIC_MULTIPLE_CHOICE,
        TYPE_STATIC,
    )
    TYPE_CHOICES_TUPLE = ((type_choice, type_choice) for type_choice in TYPE_CHOICES)

    label = LocalizedField(blank=False, null=False, required=False)
    type = models.CharField(choices=TYPE_CHOICES_TUPLE, max_length=23)
    is_required = models.TextField(default="false")
    is_hidden = models.TextField(default="false")
    is_archived = models.BooleanField(default=False)
    placeholder = LocalizedField(blank=True, null=True, required=False)
    info_text = LocalizedField(blank=True, null=True, required=False)
    static_content = LocalizedTextField(blank=True, null=True, required=False)
    configuration = JSONField(default=dict)
    meta = JSONField(default=dict)
    data_source = models.CharField(max_length=255, blank=True, null=True)
    options = models.ManyToManyField(
        "Option", through="QuestionOption", related_name="questions"
    )
    row_form = models.ForeignKey(
        Form,
        blank=True,
        null=True,
        related_name="+",
        help_text="Form that represents rows of a TableQuestion",
        on_delete=models.PROTECT,
    )
    sub_form = models.ForeignKey(
        Form,
        blank=True,
        null=True,
        related_name="+",
        help_text="Form referenced in a FormQuestion",
        on_delete=models.PROTECT,
    )

    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        help_text="Reference this question has been copied from",
        on_delete=models.SET_NULL,
    )
    format_validators = ArrayField(
        models.CharField(max_length=255), blank=True, default=list
    )

    @property
    def min_length(self):
        return self.configuration.get("min_length")

    @min_length.setter
    def min_length(self, value):
        self.configuration["min_length"] = value

    @property
    def max_length(self):
        return self.configuration.get("max_length")

    @max_length.setter
    def max_length(self, value):
        self.configuration["max_length"] = value

    @property
    def max_value(self):
        return self.configuration.get("max_value")

    @max_value.setter
    def max_value(self, value):
        self.configuration["max_value"] = value

    @property
    def min_value(self):
        return self.configuration.get("min_value")

    @min_value.setter
    def min_value(self, value):
        self.configuration["min_value"] = value

    def empty_value(self):
        """Return empty value for this question type."""

        empties = {
            Question.TYPE_MULTIPLE_CHOICE: [],
            Question.TYPE_TABLE: [],
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE: [],
        }
        return empties.get(self.type, None)

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class QuestionOption(NaturalKeyModel):
    question = models.ForeignKey("Question", on_delete=models.CASCADE)
    option = models.ForeignKey("Option", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    def natural_key(self):
        return f"{self.question_id}.{self.option_id}"

    class Meta:
        ordering = ["-sort"]
        unique_together = ("option", "question")


class Option(SlugModel):
    label = LocalizedField(blank=False, null=False, required=False)
    is_archived = models.BooleanField(default=False)
    meta = JSONField(default=dict)
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        related_name="+",
        help_text="Reference this option has been copied from",
        on_delete=models.SET_NULL,
    )

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class DocumentManager(models.Manager):
    @transaction.atomic
    def create_document_for_task(self, task, user):
        """Create a document for a given task."""
        if task.form_id is not None:
            return Document.objects.create(
                form_id=task.form_id,
                created_by_user=user.username,
                created_by_group=user.group,
            )


class Document(UUIDModel):
    objects = DocumentManager()

    family = models.ForeignKey(
        "self",
        help_text="Family id which document belongs too.",
        null=True,
        on_delete=models.CASCADE,
        related_name="+",
    )
    form = models.ForeignKey(
        "caluma_form.Form", on_delete=models.DO_NOTHING, related_name="documents"
    )
    source = models.ForeignKey(
        "self",
        blank=True,
        null=True,
        help_text="Reference this document has been copied from",
        related_name="copies",
        on_delete=models.SET_NULL,
    )
    meta = JSONField(default=dict)

    def set_family(self, root_doc):
        """Set the family to the given root_doc.

        Apply the same to all row documents within this document,
        recursively.
        """
        if self.family != root_doc:
            self.family = root_doc
            self.save()

        table_answers = Answer.objects.filter(
            document=self,
            # TODO this is not REALLY required, as only table answers
            # have child documents anyhow. Leaving it out would avoid
            # a JOIN to the question table
            question__type=Question.TYPE_TABLE,
        )

        child_documents = AnswerDocument.objects.filter(
            answer__in=table_answers
        ).select_related("document")

        for answer_document in child_documents:
            answer_document.document.set_family(root_doc)

    def copy(self, family=None):
        """Create a copy including all its answers."""

        new_document = type(self).objects.create(
            form=self.form, meta=dict(self.meta), source=self, family=family
        )
        if not family:
            family = new_document

        # copy answers
        for source_answer in self.answers.all():
            new_answer = new_document.answers.create(
                question=source_answer.question,
                value=source_answer.value,
                meta=dict(source_answer.meta),
                date=source_answer.date,
            )

            if source_answer.question.type == Question.TYPE_FILE:
                new_answer.file = source_answer.file.copy()
                new_answer.save()

            # TableAnswer: copy AnswerDocument too
            for answer_doc in AnswerDocument.objects.filter(answer=source_answer):
                new_doc = answer_doc.document.copy(family=family)

                AnswerDocument.objects.create(
                    answer=new_answer, document=new_doc, sort=answer_doc.sort
                )

        return new_document

    class Meta:
        indexes = [GinIndex(fields=["meta"])]


class Answer(UUIDModel):
    question = models.ForeignKey(
        "caluma_form.Question", on_delete=models.DO_NOTHING, related_name="answers"
    )
    value = JSONField(null=True, blank=True)
    meta = JSONField(default=dict)
    document = models.ForeignKey(
        Document, on_delete=models.CASCADE, related_name="answers"
    )
    documents = models.ManyToManyField(
        Document, through="AnswerDocument", related_name="+"
    )
    date = models.DateField(null=True, blank=True)
    file = models.OneToOneField(
        "File", on_delete=models.SET_NULL, null=True, blank=True
    )

    def delete(self, *args, **kwargs):
        if self.file:
            self.file.delete()
        super().delete(args, kwargs)

    class Meta:
        # a question may only be answerd once per document
        unique_together = ("document", "question")
        indexes = [models.Index(fields=["date"]), GinIndex(fields=["meta", "value"])]


class File(UUIDModel):
    name = models.CharField(max_length=255)

    def _move_blob(self):
        # move the file on update
        # this makes sure it stays available when querying the history
        old_file = self.history.first()
        new_name = f"{old_file.pk}_{old_file.name}"
        client.move_object(self.object_name, new_name)

    def copy(self, *args, **kwargs):
        copy = File.objects.create(name=self.name)
        client.copy_object(self.object_name, copy.object_name)
        return copy

    def delete(self, *args, **kwargs):
        self._move_blob()
        super().delete(*args, **kwargs)

    def save(self, *args, **kwargs):
        if self.created_at:
            self._move_blob()
        super().save(*args, **kwargs)

    @property
    def object_name(self):
        return f"{self.pk}_{self.name}"

    @property
    def upload_url(self):
        return client.upload_url(self.object_name)

    @property
    def download_url(self):
        return client.download_url(self.object_name)

    @property
    def metadata(self):
        return client.stat_object(self.object_name).__dict__


@receiver(post_init, sender=Document)
def set_document_family(sender, instance, **kwargs):
    """
    Ensure document has the family pointer set.

    This sets the document's family if not overruled or set already.
    The family will be in the corresponding mutations when a tree
    structure is created or restructured.
    """
    if instance.family_id is None:
        # can't use instance.set_family() here as the instance isn't
        # in DB yet, causing integrity errors
        instance.family = instance


class AnswerDocument(UUIDModel):
    answer = models.ForeignKey("Answer", on_delete=models.CASCADE)
    document = models.ForeignKey("Document", on_delete=models.CASCADE)
    sort = models.PositiveIntegerField(editable=False, db_index=True, default=0)

    class Meta:
        ordering = ["-sort"]
        unique_together = ("answer", "document")


class DynamicOption(UUIDModel):
    slug = models.CharField(max_length=255)
    label = LocalizedField(blank=False, null=False, required=False)
    document = models.ForeignKey("Document", on_delete=models.CASCADE)
    question = models.ForeignKey("Question", on_delete=models.CASCADE)

    class Meta:
        unique_together = ("slug", "document", "question")



from itertools import count

import pyjexl
from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.exceptions import ParseError
from rest_framework import exceptions


class Cache:
    """Custom cache.

    For JEXL expressions, we cannot use django's cache infrastructure, as the
    cached objects are pickled. This won't work for parsed JEXL expressions, as
    they contain lambdas etc.
    """

    def __init__(self, max_size=2000, evict_to=1500):
        self.max_size = max_size
        self.evict_to = evict_to

        self._cache = {}
        self._mru = {}
        self._mru_count = count()

    def get_or_set(self, key, default):
        if key in self._cache:
            self._mru[key] = next(self._mru_count)
            return self._cache[key]

        ret = self._cache[key] = default()
        self._mru[key] = next(self._mru_count)

        if len(self._mru) > self.max_size:
            self._evict()

        return ret

    def _evict(self):
        to_purge = list(sorted(self._cache.keys(), key=lambda key: self._mru[key]))
        # to_purge contains all keys, but we only want to evict the oldest
        # ones
        num_to_evict = len(to_purge) - self.evict_to

        for key in to_purge[:num_to_evict]:
            del self._cache[key]
            del self._mru[key]


class JexlValidator(object):
    def __init__(self, jexl):
        self.jexl = jexl

    def __call__(self, value):
        errors = list(self.jexl.validate(value))
        if errors:
            raise exceptions.ValidationError(errors)


class JEXL(pyjexl.JEXL):
    expr_cache = Cache()

    def parse(self, expression):
        parsed_expression = self.expr_cache.get_or_set(
            expression, lambda: super(JEXL, self).parse(expression)
        )
        return parsed_expression

    def validate(self, expression, ValidatingAnalyzerClass=ValidatingAnalyzer):
        try:
            for res in self.analyze(expression, ValidatingAnalyzerClass):
                yield res
        except ParseError as err:
            yield str(err)


class ExtractTransformSubjectAnalyzer(ValidatingAnalyzer):
    """
    Extract all subject values of given transforms.

    If no transforms are given all subjects of all transforms will be extracted.
    """

    def __init__(self, config, transforms=None):
        self.transforms = transforms if transforms else []
        super().__init__(config)

    def visit_Transform(self, transform):
        if not self.transforms or transform.name in self.transforms:
            # can only extract subject's value if subject is not a transform itself
            if not isinstance(transform.subject, type(transform)):
                yield transform.subject.value
        yield from self.generic_visit(transform)



import itertools

from django.db import transaction
from django.utils import timezone
from rest_framework import exceptions
from simple_history.utils import bulk_create_with_history

from ..caluma_core import serializers
from ..caluma_form.models import Document, Form
from . import models, validators
from .jexl import FlowJexl, GroupJexl


def evaluate_assigned_groups(task):
    if task.address_groups:
        return GroupJexl().evaluate(task.address_groups)

    return []


def get_addressed_groups(task):
    addressed_groups = [evaluate_assigned_groups(task)]
    if task.is_multiple_instance:
        addressed_groups = [[x] for x in addressed_groups[0]]
    return addressed_groups


class FlowJexlField(serializers.JexlField):
    def __init__(self, **kwargs):
        super().__init__(FlowJexl(), **kwargs)


class GroupJexlField(serializers.JexlField):
    def __init__(self, **kwargs):
        super().__init__(GroupJexl(), **kwargs)


class SaveWorkflowSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Workflow
        fields = (
            "slug",
            "name",
            "description",
            "meta",
            "start_tasks",
            "allow_all_forms",
            "allow_forms",
            "is_archived",
            "is_published",
        )


class AddWorkflowFlowSerializer(serializers.ModelSerializer):
    workflow = serializers.GlobalIDField(source="slug")
    tasks = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Task.objects, many=True
    )
    next = FlowJexlField(required=True)

    def validate_next(self, value):
        jexl = FlowJexl()
        tasks = set(jexl.extract_tasks(value))

        if not tasks:
            raise exceptions.ValidationError(
                f"jexl `{value}` does not contain any tasks as return value"
            )

        available_tasks = set(
            models.Task.objects.filter(slug__in=tasks).values_list("slug", flat=True)
        )

        not_found_tasks = tasks - available_tasks
        if not_found_tasks:
            raise exceptions.ValidationError(
                f"jexl `{value}` contains invalid tasks [{', '.join(not_found_tasks)}]"
            )

        return value

    @transaction.atomic
    def update(self, instance, validated_data):
        user = self.context["request"].user
        tasks = validated_data["tasks"]
        models.Flow.objects.filter(task_flows__task__in=tasks).delete()
        flow = models.Flow.objects.create(
            next=validated_data["next"],
            created_by_user=user.username,
            created_by_group=user.group,
        )
        task_flows = [
            models.TaskFlow(task=task, workflow=instance, flow=flow) for task in tasks
        ]
        bulk_create_with_history(task_flows, models.TaskFlow)

        return instance

    class Meta:
        fields = ("workflow", "tasks", "next")
        model = models.Workflow


class RemoveFlowSerializer(serializers.ModelSerializer):
    flow = serializers.GlobalIDField(source="id")

    def update(self, instance, validated_data):
        models.Flow.objects.filter(pk=instance.pk).delete()
        return instance

    class Meta:
        fields = ("flow",)
        model = models.Flow


class SaveTaskSerializer(serializers.ModelSerializer):
    address_groups = GroupJexlField(
        required=False,
        allow_null=True,
        help_text=models.Task._meta.get_field("address_groups").help_text,
    )

    class Meta:
        model = models.Task
        fields = (
            "slug",
            "name",
            "description",
            "meta",
            "address_groups",
            "is_archived",
            "lead_time",
            "is_multiple_instance",
        )


class SaveSimpleTaskSerializer(SaveTaskSerializer):
    def validate(self, data):
        data["type"] = models.Task.TYPE_SIMPLE
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        pass


class SaveCompleteWorkflowFormTaskSerializer(SaveTaskSerializer):
    def validate(self, data):
        data["type"] = models.Task.TYPE_COMPLETE_WORKFLOW_FORM
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        pass


class SaveCompleteTaskFormTaskSerializer(SaveTaskSerializer):
    form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=Form.objects, required=True
    )

    def validate(self, data):
        data["type"] = models.Task.TYPE_COMPLETE_TASK_FORM
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        fields = SaveTaskSerializer.Meta.fields + ("form",)


class CaseSerializer(serializers.ModelSerializer):
    workflow = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Workflow.objects.prefetch_related("start_tasks")
    )
    parent_work_item = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.WorkItem.objects, required=False
    )
    form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=Form.objects, required=False
    )

    def validate(self, data):
        form = data.get("form")
        workflow = data["workflow"]

        if form:
            if (
                not workflow.allow_all_forms
                and not workflow.allow_forms.filter(pk=form.pk).exists()
            ):
                raise exceptions.ValidationError(
                    f"Workflow {workflow.pk} does not allow to start case with form {form.pk}"
                )

        return super().validate(data)

    @transaction.atomic
    def create(self, validated_data):
        user = self.context["request"].user
        parent_work_item = validated_data.get("parent_work_item")
        validated_data["status"] = models.Case.STATUS_RUNNING

        form = validated_data.pop("form", None)
        if form:
            validated_data["document"] = Document.objects.create(
                form=form, created_by_user=user.username, created_by_group=user.group
            )

        instance = super().create(validated_data)
        if parent_work_item:
            parent_work_item.child_case = instance
            parent_work_item.save()

        workflow = instance.workflow
        tasks = workflow.start_tasks.all()

        work_items = itertools.chain(
            *[
                [
                    models.WorkItem(
                        addressed_groups=groups,
                        task_id=task.pk,
                        deadline=task.calculate_deadline(),
                        document=Document.objects.create_document_for_task(task, user),
                        case=instance,
                        status=models.WorkItem.STATUS_READY,
                        created_by_user=user.username,
                        created_by_group=user.group,
                    )
                    for groups in get_addressed_groups(task)
                ]
                for task in tasks
            ]
        )

        bulk_create_with_history(work_items, models.WorkItem)
        return instance

    class Meta:
        model = models.Case
        fields = ("workflow", "meta", "parent_work_item", "form")


class SaveCaseSerializer(CaseSerializer):
    class Meta(CaseSerializer.Meta):
        fields = ("id", "workflow", "meta", "parent_work_item", "form")


class CancelCaseSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    class Meta:
        model = models.Case
        fields = ("id",)

    def validate(self, data):
        if self.instance.status != models.Case.STATUS_RUNNING:
            raise exceptions.ValidationError("Only running cases can be canceled.")

        user = self.context["request"].user
        data["status"] = models.Case.STATUS_CANCELED
        data["closed_at"] = timezone.now()
        data["closed_by_user"] = user.username
        data["closed_by_group"] = user.group
        return super().validate(data)

    @transaction.atomic
    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)
        user = self.context["request"].user
        instance.work_items.exclude(
            status__in=[
                models.WorkItem.STATUS_COMPLETED,
                models.WorkItem.STATUS_CANCELED,
            ]
        ).update(
            status=models.WorkItem.STATUS_CANCELED,
            closed_at=timezone.now(),
            closed_by_user=user.username,
            closed_by_group=user.group,
        )
        return instance


class CompleteWorkItemSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    def validate(self, data):
        user = self.context["request"].user
        validators.WorkItemValidator().validate(
            status=self.instance.status,
            child_case=self.instance.child_case,
            task=self.instance.task,
            case=self.instance.case,
            document=self.instance.document,
            info=self.context["info"],
        )
        data["status"] = models.WorkItem.STATUS_COMPLETED
        data["closed_at"] = timezone.now()
        data["closed_by_user"] = user.username
        data["closed_by_group"] = user.group
        return super().validate(data)

    def _can_continue(self, instance, task):
        # If a "multiple instance" task has running siblings, the task is not completed
        if task.is_multiple_instance:
            return not instance.case.work_items.filter(
                task=task, status=models.WorkItem.STATUS_READY
            ).exists()
        return instance.case.work_items.filter(
            task=task,
            status__in=(
                models.WorkItem.STATUS_COMPLETED,
                models.WorkItem.STATUS_SKIPPED,
            ),
        ).exists()

    @transaction.atomic
    def update(self, instance, validated_data):

        instance = super().update(instance, validated_data)
        user = self.context["request"].user
        case = instance.case

        if not self._can_continue(instance, instance.task):
            return instance

        flow = models.Flow.objects.filter(task_flows__task=instance.task_id).first()
        flow_referenced_tasks = models.Task.objects.filter(task_flows__flow=flow)

        all_complete = all(
            self._can_continue(instance, task) for task in flow_referenced_tasks
        )

        if flow and all_complete:
            jexl = FlowJexl()
            result = jexl.evaluate(flow.next)
            if not isinstance(result, list):
                result = [result]

            tasks = models.Task.objects.filter(pk__in=result)

            work_items = itertools.chain(
                *[
                    [
                        models.WorkItem(
                            addressed_groups=groups,
                            task_id=task.pk,
                            deadline=task.calculate_deadline(),
                            document=Document.objects.create_document_for_task(
                                task, user
                            ),
                            case=case,
                            status=models.WorkItem.STATUS_READY,
                            created_by_user=user.username,
                            created_by_group=user.group,
                        )
                        for groups in get_addressed_groups(task)
                    ]
                    for task in tasks
                ]
            )

            bulk_create_with_history(work_items, models.WorkItem)
        else:
            # no more tasks, mark case as complete
            case.status = models.Case.STATUS_COMPLETED
            case.closed_at = timezone.now()
            case.closed_by_user = user.username
            case.closed_by_group = user.group
            case.save()

        return instance

    class Meta:
        model = models.WorkItem
        fields = ("id",)


class SkipWorkItemSerializer(CompleteWorkItemSerializer):
    def validate(self, data):
        if self.instance.status != models.WorkItem.STATUS_READY:
            raise exceptions.ValidationError("Only READY work items can be skipped")

        user = self.context["request"].user
        data["status"] = models.WorkItem.STATUS_SKIPPED
        data["closed_at"] = timezone.now()
        data["closed_by_user"] = user.username
        data["closed_by_group"] = user.group
        # We skip parent validation, as the work item is now "skipped",
        # meaning no other conditions need apply

        return data


class SaveWorkItemSerializer(serializers.ModelSerializer):
    work_item = serializers.GlobalIDField(source="id")

    class Meta:
        model = models.WorkItem
        fields = ("work_item", "assigned_users", "deadline", "meta")


class CreateWorkItemSerializer(serializers.ModelSerializer):
    case = serializers.GlobalIDPrimaryKeyRelatedField(queryset=models.Case.objects)
    multiple_instance_task = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Task.objects, source="task"
    )

    def validate_multiple_instance_task(self, task):
        if not task.is_multiple_instance:
            raise exceptions.ValidationError(
                f"The given task type {task.type} does not allow creating multiple instances of it. Please set `isMultipleInstance` to true."
            )
        return task

    def validate(self, data):
        user = self.context["request"].user
        case = data["case"]
        task = data["task"]

        if not case.work_items.filter(
            task=task, status=models.WorkItem.STATUS_READY
        ).exists():
            raise exceptions.ValidationError(
                f"The given case {case.pk} does not have any running work items corresponding to the task {task.pk}. A new instance of a `MultipleInstanceTask` can only be created when there is at least one running sibling work item."
            )

        data["document"] = Document.objects.create_document_for_task(task, user)
        data["status"] = models.WorkItem.STATUS_READY
        return super().validate(data)

    class Meta:
        model = models.WorkItem
        fields = (
            "case",
            "multiple_instance_task",
            "assigned_users",
            "addressed_groups",
            "deadline",
            "meta",
        )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_remove_answer[integer-23] 1"] = {
    "removeAnswer": {
        "answer": {
            "__typename": "IntegerAnswer",
            "id": "SW50ZWdlckFuc3dlcjpOb25l",
            "meta": {},
        },
        "clientMutationId": None,
    }
}



from django.conf import settings
from django.conf.urls import url

from .caluma_user.views import AuthenticationGraphQLView

urlpatterns = [
    url(
        r"^graphql",
        AuthenticationGraphQLView.as_view(graphiql=settings.DEBUG),
        name="graphql",
    )
]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.17 on 2018-12-18 13:52
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("caluma_form", "0001_initial"),
        ("caluma_workflow", "0003_auto_20181217_1051"),
    ]

    operations = [
        migrations.AddField(
            model_name="task",
            name="form",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="tasks",
                to="caluma_form.Form",
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="document",
            field=models.OneToOneField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="work_item",
                to="caluma_form.Document",
            ),
        ),
        migrations.AlterField(
            model_name="task",
            name="type",
            field=models.CharField(
                choices=[
                    ("simple", "Task which can only be marked as completed."),
                    (
                        "complete_workflow_form",
                        "Task completing defined workflow form.",
                    ),
                    ("complete_task_form", "Task completing defined task form."),
                ],
                max_length=50,
            ),
        ),
    ]



import pytest

from .. import models


@pytest.mark.parametrize(
    "question__type,answer__value", [(models.Question.TYPE_INTEGER, 23)]
)
def test_remove_answer(db, snapshot, question, answer, schema_executor):
    query = """
        mutation RemoveAnswer($input: RemoveAnswerInput!) {
          removeAnswer(input: $input) {
            answer {
              id
              meta
              __typename
            }
            clientMutationId
          }
        }
    """

    result = schema_executor(query, variables={"input": {"answer": str(answer.pk)}})

    assert not result.errors
    with pytest.raises(models.Answer.DoesNotExist):
        answer.refresh_from_db()

    snapshot.assert_match(result.data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-01-30 09:20
from __future__ import unicode_literals

import uuid

import django.contrib.postgres.fields.jsonb
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0002_auto_20181221_1517")]

    operations = [
        migrations.CreateModel(
            name="AnswerDocument",
            fields=[
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("modified_at", models.DateTimeField(auto_now=True)),
                (
                    "created_by_user",
                    models.CharField(blank=True, max_length=150, null=True),
                ),
                (
                    "created_by_group",
                    models.CharField(
                        blank=True, db_index=True, max_length=150, null=True
                    ),
                ),
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "sort",
                    models.PositiveIntegerField(
                        db_index=True, default=0, editable=False
                    ),
                ),
            ],
            options={"ordering": ("-sort", "id")},
        ),
        migrations.AddField(
            model_name="document",
            name="family",
            field=models.UUIDField(
                db_index=True,
                default=uuid.uuid4,
                help_text="Family id which document belongs too.",
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="question",
            name="row_form",
            field=models.ForeignKey(
                blank=True,
                help_text="One row of table is represented by this form",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AlterField(
            model_name="answer",
            name="value",
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("checkbox", "checkbox"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("radio", "radio"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                ],
                max_length=10,
            ),
        ),
        migrations.AddField(
            model_name="answerdocument",
            name="answer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="caluma_form.Answer"
            ),
        ),
        migrations.AddField(
            model_name="answerdocument",
            name="document",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="caluma_form.Document"
            ),
        ),
        migrations.AddField(
            model_name="answer",
            name="documents",
            field=models.ManyToManyField(
                related_name="_answer_documents_+",
                through="caluma_form.AnswerDocument",
                to="caluma_form.Document",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="answerdocument", unique_together=set([("answer", "document")])
        ),
    ]



from warnings import warn

from django.conf import settings


class BaseUser:  # pragma: no cover
    def __init__(self):
        self.username = None
        self.groups = []
        self.token = None
        self.claims = {}
        self.is_authenticated = False

    @property
    def group(self):
        raise NotImplementedError

    def __getattribute__(self, name):
        if name in ["userinfo", "introspection"]:
            warn(
                f'"{name}" is deprecated. Use the new "claims" attribute',
                DeprecationWarning,
            )
            return self.claims
        return super().__getattribute__(name)

    def __str__(self):
        raise NotImplementedError


class AnonymousUser(BaseUser):
    @property
    def group(self):
        return None

    def __str__(self):
        return "AnonymousUser"


class OIDCUser(BaseUser):
    def __init__(self, token: str, userinfo: dict = None, introspection: dict = None):
        super().__init__()

        self.claims, self.claims_source = self._get_claims(userinfo, introspection)
        self.username = self.claims[settings.OIDC_USERNAME_CLAIM]
        self.groups = self.claims.get(settings.OIDC_GROUPS_CLAIM)
        self.token = token
        self.is_authenticated = True

    def _get_claims(self, userinfo, introspection):
        result = (userinfo, "userinfo")
        if all([userinfo, introspection]):  # pragma: no cover
            raise AttributeError("You can't set userinfo AND introspection.")
        elif not any([userinfo, introspection]):  # pragma: no cover
            raise AttributeError("You must provide either userinfo or introspection.")
        elif introspection is not None:
            result = (introspection, "introspection")
        return result

    @property
    def group(self):
        return self.groups[0] if self.groups else None

    def __str__(self):
        return self.username



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-03-21 17:22
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0008_auto_20190319_1720")]

    operations = [
        migrations.AddField(
            model_name="answer",
            name="value_document",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name="parent_answers",
                to="caluma_form.Document",
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="sub_form",
            field=models.ForeignKey(
                blank=True,
                help_text="Form referenced in a FormQuestion",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="row_form",
            field=models.ForeignKey(
                blank=True,
                help_text="Form that represents rows of a TableQuestion",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("date", "date"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                    ("form", "form"),
                ],
                max_length=15,
            ),
        ),
    ]



from graphql_relay import from_global_id


def extract_global_id(id):
    """
    Extract id from global id throwing away type.

    In case it is not a base64 encoded value it will return id as is.
    This way it can be used as input type by simply defining id without its specific type.
    """

    try:
        _, result = from_global_id(id)
    except ValueError:
        result = id

    return result



import pytest
from django.core.exceptions import ImproperlyConfigured

from .. import models, serializers
from ..mutation import Mutation
from ..validations import BaseValidation, validation_for
from .fake_model import get_fake_model


def test_custom_validation_override_validate_custom_node(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomValidation(BaseValidation):
        @validation_for(CustomMutation)
        def validate_custom_mutation(self, mutation, data, info):
            data["test"] = "test"
            return data

    data = CustomValidation().validate(CustomMutation, {}, info)
    assert data["test"] == "test"


def test_custom_validation_override_validate_with_duplicates(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomValidation(BaseValidation):
        @validation_for(CustomMutation)
        def validate_custom_mutation(self, mutation, data, info):  # pragma: no cover
            return data

        @validation_for(CustomMutation)
        def validate_custom_mutation_2(self, mutation, data, info):  # pragma: no cover
            return data

    with pytest.raises(ImproperlyConfigured):
        CustomValidation()


def test_custom_validation_override_validate_no_class_match(db, info):
    FakeModel = get_fake_model(model_base=models.UUIDModel)

    class Serializer(serializers.ModelSerializer):
        class Meta:
            model = FakeModel
            fields = "__all__"

    class CustomMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class AnotherMutation(Mutation):
        class Meta:
            serializer_class = Serializer

    class CustomValidation(BaseValidation):
        @validation_for(CustomMutation)
        def validate_custom_mutation(self, mutation, data, info):  # pragma: no cover
            return data

    assert CustomValidation().validate(AnotherMutation, {}, info) == {}



import pytest

from .. import structure, validators
from ..jexl import QuestionJexl, QuestionMissing
from ..models import Question


@pytest.mark.parametrize(
    "expression,num_errors",
    [
        # correct case
        ('"question-slug"|answer|mapby', 0),
        # invalid subject type
        ("100|answer", 1),
        # two invalid subject types
        ('["test"]|answer || 1.0|answer', 2),
        # invalid operator
        ("'question-slug1'|answer &&& 'question-slug2'|answer", 1),
    ],
)
def test_question_jexl_validate(expression, num_errors):
    jexl = QuestionJexl()
    assert len(list(jexl.validate(expression))) == num_errors


@pytest.mark.parametrize(
    "expression,result",
    [
        ("[1,2] intersects [2,3]", True),
        ("[1,2] intersects [3,4]", False),
        ("[] intersects []", False),
        ("[1] intersects []", False),
        ("['foo'] intersects ['bar', 'bazz']", False),
        ("['foo'] intersects ['foo', 'foo']", True),
        ("[1] intersects [1] && [2] intersects [2]", True),
        ("[2] intersects [1] + [2]", True),
    ],
)
def test_intersects_operator(expression, result):
    assert QuestionJexl().evaluate(expression) == result


@pytest.mark.parametrize("form__slug", ["f-main-slug"])
def test_jexl_form(db, form):
    answer_by_question = {
        "a1": {"value": "A1", "form": form},
        "b1": {"value": "B1", "form": form},
    }

    assert (
        QuestionJexl({"answers": answer_by_question, "form": form}).evaluate("form")
        == "f-main-slug"
    )


def test_all_deps_hidden(db, form, document_factory, form_question_factory):
    q1 = form_question_factory(form=form, question__is_hidden="true").question
    q2 = form_question_factory(
        form=form,
        question__is_hidden=f"'{q1.slug}'|answer=='blah'",
        question__is_required=f"'{q1.slug}'|answer=='blah'",
    ).question
    document = document_factory(form=form)

    qj = QuestionJexl(
        {
            "document": document,
            "answers": {},
            "form": form,
            "structure": structure.FieldSet(document, document.form),
        }
    )
    assert qj.is_hidden(q2)
    assert not qj.is_required(q2)


@pytest.mark.parametrize("fq_is_hidden", ["true", "false"])
def test_indirectly_hidden_dependency(
    db,
    form_question_factory,
    form_factory,
    question_factory,
    document_factory,
    answer_factory,
    info,
    fq_is_hidden,
):
    # Questions can not only be hidden by evaluating their is_hidden expression
    # but also if they're part of a subform, where the containing form question
    # is hidden.
    #
    # Showcase form to demonstrate this (f: form, q:question, d:document, a:answer):
    # f:topform
    #   q:formquestion - hidden=true
    #      f:subform
    #         q:subquestion - hidden=false
    #   q:depquestion - required = subquestion|answer=='blah'
    #
    # d:topdoc f=topform
    #    a:subanswer q=subquestion, value='blah'
    #    (depquestion has no answer, as it's indirectly not required)
    #
    # Note: The above structure shows the actual "feature" case.
    # We also test with `subquestion` not hidden, to ensure both ways
    # work correctly

    # Build the form first...
    topform = form_factory(slug="top")
    subform = form_factory(slug="subform")

    formquestion = question_factory(
        type=Question.TYPE_FORM, is_hidden=fq_is_hidden, sub_form=subform
    )
    subquestion = question_factory(
        type=Question.TYPE_INTEGER, is_hidden="false", slug="subquestion"
    )
    depquestion = question_factory(
        type=Question.TYPE_INTEGER, is_required="'subquestion'|answer=='blah'"
    )
    form_question_factory(form=topform, question=formquestion)
    form_question_factory(form=topform, question=depquestion)

    form_question_factory(form=subform, question=subquestion)

    # ... then build the document
    topdoc = document_factory(form=topform)
    answer_factory(document=topdoc, question=subquestion, value="blah")

    validator = validators.DocumentValidator()

    if fq_is_hidden == "true":
        # parent is hidden, required question cannot be
        # shown, thus is implicitly hidden
        validator.validate(topdoc, info)
        assert True  # above did not fail
    else:
        with pytest.raises(validators.CustomValidationError):
            validator.validate(topdoc, info)


def test_reference_missing_question(
    db, form_question_factory, form_factory, question_factory, document_factory, info
):
    topform = form_factory(slug="top")

    depquestion = question_factory(
        type=Question.TYPE_INTEGER, is_hidden="'subquestion-missing'|answer=='blah'"
    )
    form_question_factory(form=topform, question=depquestion)

    # ... then build the document
    topdoc = document_factory(form=topform)
    validator = validators.DocumentValidator()

    with pytest.raises(QuestionMissing):
        validator.validate(topdoc, info)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots[
    "test_query_all_questions[integer-question__configuration0-None-question__format_validators0] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "IntegerQuestion",
                    "id": "SW50ZWdlclF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "integerMaxValue": 10,
                    "integerMinValue": 0,
                    "label": "Brian Williams",
                    "meta": {},
                    "placeholder": "",
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[float-question__configuration1-None-question__format_validators1] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "FloatQuestion",
                    "floatMaxValue": 1.0,
                    "floatMinValue": 0.0,
                    "id": "RmxvYXRRdWVzdGlvbjplZmZvcnQtbWVldA==",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "placeholder": "",
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[float-question__configuration2-None-question__format_validators2] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "FloatQuestion",
                    "floatMaxValue": None,
                    "floatMinValue": None,
                    "id": "RmxvYXRRdWVzdGlvbjplZmZvcnQtbWVldA==",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "placeholder": "",
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[date-question__configuration3-None-question__format_validators3] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "DateQuestion",
                    "id": "RGF0ZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[text-question__configuration4-None-question__format_validators4] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "TextQuestion",
                    "formatValidators": {
                        "edges": [
                            {
                                "node": {
                                    "errorMsg": "Please enter a valid Email address.",
                                    "name": "E-mail",
                                    "regex": "(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\\.[a-zA-Z0-9-.]+$)",
                                    "slug": "email",
                                }
                            }
                        ]
                    },
                    "id": "VGV4dFF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "label": "Brian Williams",
                    "maxLength": None,
                    "meta": {},
                    "minLength": 10,
                    "placeholder": "",
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[textarea-question__configuration5-None-question__format_validators5] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "TextareaQuestion",
                    "formatValidators": {"edges": []},
                    "id": "VGV4dGFyZWFRdWVzdGlvbjplZmZvcnQtbWVldA==",
                    "infoText": "",
                    "label": "Brian Williams",
                    "maxLength": 10,
                    "meta": {},
                    "minLength": None,
                    "placeholder": "",
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[choice-question__configuration6-None-question__format_validators6] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "ChoiceQuestion",
                    "id": "Q2hvaWNlUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "options": {
                        "edges": [{"node": {"slug": "treatment-radio"}}],
                        "totalCount": 1,
                    },
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[multiple_choice-question__configuration7-None-question__format_validators7] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "MultipleChoiceQuestion",
                    "id": "TXVsdGlwbGVDaG9pY2VRdWVzdGlvbjplZmZvcnQtbWVldA==",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "options": {
                        "edges": [{"node": {"slug": "treatment-radio"}}],
                        "totalCount": 1,
                    },
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[form-question__configuration8-None-question__format_validators8] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "FormQuestion",
                    "id": "Rm9ybVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "slug": "effort-meet",
                    "subForm": {"slug": "suggest-traditional"},
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[file-question__configuration9-None-question__format_validators9] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "FileQuestion",
                    "id": "RmlsZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[dynamic_choice-question__configuration10-MyDataSource-question__format_validators10] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "DynamicChoiceQuestion",
                    "id": "RHluYW1pY0Nob2ljZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "options": {
                        "edges": [
                            {"node": {"label": "1", "slug": "1"}},
                            {"node": {"label": "5.5", "slug": "5.5"}},
                            {"node": {"label": "sdkj", "slug": "sdkj"}},
                            {"node": {"label": "info", "slug": "value"}},
                            {"node": {"label": "something", "slug": "something"}},
                            {
                                "node": {
                                    "label": "english description",
                                    "slug": "translated_value",
                                }
                            },
                        ]
                    },
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[dynamic_multiple_choice-question__configuration11-MyDataSource-question__format_validators11] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "DynamicMultipleChoiceQuestion",
                    "id": "RHluYW1pY011bHRpcGxlQ2hvaWNlUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "options": {
                        "edges": [
                            {"node": {"label": "1", "slug": "1"}},
                            {"node": {"label": "5.5", "slug": "5.5"}},
                            {"node": {"label": "sdkj", "slug": "sdkj"}},
                            {"node": {"label": "info", "slug": "value"}},
                            {"node": {"label": "something", "slug": "something"}},
                            {
                                "node": {
                                    "label": "english description",
                                    "slug": "translated_value",
                                }
                            },
                        ]
                    },
                    "slug": "effort-meet",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots[
    "test_query_all_questions[static-question__configuration12-None-question__format_validators12] 1"
] = {
    "allQuestions": {
        "edges": [
            {
                "node": {
                    "__typename": "StaticQuestion",
                    "id": "U3RhdGljUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
                    "infoText": "",
                    "label": "Brian Williams",
                    "meta": {},
                    "slug": "effort-meet",
                    "staticContent": """Line whatever team suggest traditional boy. Drop argue move. Anyone remember prove.
Kid avoid player relationship to range whose. Draw free property consider.""",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots["test_save_question[true-True-SaveTextQuestion] 1"] = {
    "saveTextQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "TextQuestion",
            "id": "VGV4dFF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_question[true-True-SaveTextareaQuestion] 1"] = {
    "saveTextareaQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "TextareaQuestion",
            "id": "VGV4dGFyZWFRdWVzdGlvbjplZmZvcnQtbWVldA==",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_question[true-True-SaveIntegerQuestion] 1"] = {
    "saveIntegerQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "IntegerQuestion",
            "id": "SW50ZWdlclF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_question[true-True-SaveFloatQuestion] 1"] = {
    "saveFloatQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "FloatQuestion",
            "id": "RmxvYXRRdWVzdGlvbjplZmZvcnQtbWVldA==",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_question[true-True-SaveDateQuestion] 1"] = {
    "saveDateQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DateQuestion",
            "id": "RGF0ZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_question[true-True-SaveFileQuestion] 1"] = {
    "saveFileQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "FileQuestion",
            "id": "RmlsZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_float_question[float-question__configuration0-True] 1"] = {
    "saveFloatQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "FloatQuestion",
            "id": "RmxvYXRRdWVzdGlvbjplZmZvcnQtbWVldA==",
            "label": "Brian Williams",
            "maxValue": 10.0,
            "meta": {},
            "minValue": 0.0,
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_integer_question[integer-question__configuration0-True] 1"] = {
    "saveIntegerQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "IntegerQuestion",
            "id": "SW50ZWdlclF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_multiple_choice_question[multiple_choice] 1"] = {
    "saveMultipleChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "MultipleChoiceQuestion",
            "id": "TXVsdGlwbGVDaG9pY2VRdWVzdGlvbjplZmZvcnQtbWVldA==",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {
                        "node": {
                            "label": "Mariah Reynolds",
                            "slug": "provide-beyond-seek",
                        }
                    },
                    {"node": {"label": "Daniel Mann", "slug": "live-by-itself"}},
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_choice_question[choice] 1"] = {
    "saveChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "ChoiceQuestion",
            "id": "Q2hvaWNlUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "John Thomas", "slug": "suggest-traditional"}}
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_dynamic_choice_question[dynamic_choice-True] 1"] = {
    "saveDynamicChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicChoiceQuestion",
            "id": "RHluYW1pY0Nob2ljZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_dynamic_choice_question[dynamic_choice-False] 1"] = {
    "saveDynamicChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicChoiceQuestion",
            "id": "RHluYW1pY0Nob2ljZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_dynamic_choice_question[dynamic_multiple_choice-True] 1"] = {
    "saveDynamicChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicChoiceQuestion",
            "id": "RHluYW1pY0Nob2ljZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_dynamic_choice_question[dynamic_multiple_choice-False] 1"] = {
    "saveDynamicChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicChoiceQuestion",
            "id": "RHluYW1pY0Nob2ljZVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots[
    "test_save_dynamic_multiple_choice_question[dynamic_multiple_choice-True] 1"
] = {
    "saveDynamicMultipleChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicMultipleChoiceQuestion",
            "id": "RHluYW1pY011bHRpcGxlQ2hvaWNlUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots[
    "test_save_dynamic_multiple_choice_question[dynamic_multiple_choice-False] 1"
] = {
    "saveDynamicMultipleChoiceQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "DynamicMultipleChoiceQuestion",
            "id": "RHluYW1pY011bHRpcGxlQ2hvaWNlUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
            "label": "Brian Williams",
            "meta": {},
            "options": {
                "edges": [
                    {"node": {"label": "1", "slug": "1"}},
                    {"node": {"label": "5.5", "slug": "5.5"}},
                    {"node": {"label": "sdkj", "slug": "sdkj"}},
                    {"node": {"label": "info", "slug": "value"}},
                    {"node": {"label": "something", "slug": "something"}},
                    {
                        "node": {
                            "label": "english description",
                            "slug": "translated_value",
                        }
                    },
                ]
            },
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_table_question[table] 1"] = {
    "saveTableQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "TableQuestion",
            "id": "VGFibGVRdWVzdGlvbjplZmZvcnQtbWVldA==",
            "label": "Brian Williams",
            "meta": {},
            "rowForm": {"slug": "suggest-traditional"},
            "slug": "effort-meet",
        },
    }
}

snapshots["test_save_form_question[form] 1"] = {
    "saveFormQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "FormQuestion",
            "id": "Rm9ybVF1ZXN0aW9uOmVmZm9ydC1tZWV0",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
            "subForm": {"slug": "suggest-traditional"},
        },
    }
}

snapshots["test_save_static_question[static] 1"] = {
    "saveStaticQuestion": {
        "clientMutationId": "testid",
        "question": {
            "__typename": "StaticQuestion",
            "id": "U3RhdGljUXVlc3Rpb246ZWZmb3J0LW1lZXQ=",
            "label": "Brian Williams",
            "meta": {},
            "slug": "effort-meet",
            "staticContent": """Line whatever team suggest traditional boy. Drop argue move. Anyone remember prove.
Kid avoid player relationship to range whose. Draw free property consider.""",
        },
    }
}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-07 14:05
from __future__ import unicode_literals

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0003_auto_20190130_0920")]

    operations = [
        migrations.AddField(
            model_name="form",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this form has been copied from",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Form",
            ),
        ),
        migrations.AddField(
            model_name="option",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this option has been copied from",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Option",
            ),
        ),
        migrations.AddField(
            model_name="question",
            name="source",
            field=models.ForeignKey(
                blank=True,
                help_text="Reference this question has been copied from",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="caluma_form.Question",
            ),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-08 12:32
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0006_choice_type_conversion")]

    operations = [
        migrations.AlterModelOptions(
            name="answerdocument", options={"ordering": ["-sort"]}
        ),
        migrations.AlterModelOptions(
            name="formquestion", options={"ordering": ["-sort"]}
        ),
        migrations.AlterModelOptions(
            name="questionoption", options={"ordering": ["-sort"]}
        ),
    ]



from time import sleep

from django.core.cache import cache

from .data_sources import MyDataSource


def test_set_cache(info):
    cache.clear()
    ds = MyDataSource()
    result = ds.get_data_test_string(info)
    assert result == "test string"
    assert cache.get("data_source_MyDataSource") == "test string"


def test_get_from_cache(info):
    cache.clear()
    ds = MyDataSource()
    ds.get_data_uuid(info)
    cached_result = cache.get("data_source_MyDataSource")
    new_result = ds.get_data_uuid(info)
    assert cached_result == new_result


def test_expired_cache(info):
    cache.clear()
    ds = MyDataSource()
    ds.get_data_expire(info)
    cached_result = cache.get("data_source_MyDataSource")

    sleep(1.1)
    new_result = ds.get_data_uuid(info)

    assert not cached_result == new_result



import pytest

from caluma.caluma_form.models import Question

from ..relay import extract_global_id


@pytest.mark.parametrize("question__type", ["text"])
def test_inversible_with_vars(db, schema_executor, document_factory, question):

    doc_a, doc_b, doc_c = (
        document_factory(meta={"foo": "a", "blah": "xxx"}),
        document_factory(meta={"foo": "b", "blah": "xxx"}),
        document_factory(meta={"foo": "c", "blah": "blub"}),
    )

    doc_a.answers.create(question=question, value="44")
    doc_b.answers.create(question=question, value="33")
    doc_c.answers.create(question=question, value="22")

    query = """
        query test_ordering($order: [DocumentOrderSetType]) {
          allDocuments(order: $order) {
            edges {
              node {
                id
              }
            }
          }
        }
    """

    doc_a_id = str(doc_a.id)
    doc_b_id = str(doc_b.id)
    doc_c_id = str(doc_c.id)

    def result_ids(result):
        return [
            extract_global_id(n["node"]["id"])
            for n in result.data["allDocuments"]["edges"]
        ]

    def get_ordered(ordering):
        result = schema_executor(query, variables={"order": ordering})
        assert not result.errors
        return result_ids(result)

    # ordering by answer value, forward.
    ordered = get_ordered([{"answerValue": question.slug}])
    assert ordered == [doc_c_id, doc_b_id, doc_a_id]

    # ordering by answer value, backward.
    ordered = get_ordered([{"answerValue": question.slug, "direction": "DESC"}])
    assert ordered == [doc_a_id, doc_b_id, doc_c_id]

    # order by meta key "blah" first, which contains the same value twice,
    # then by the answer value, which orders the ambiguity between
    # the first two same-value documents
    ordered = get_ordered(
        [{"meta": "blah"}, {"answerValue": question.slug, "direction": "DESC"}]
    )
    assert ordered == [doc_c_id, doc_a_id, doc_b_id]

    # same, but secondary order goes backwards
    ordered = get_ordered([{"meta": "blah"}, {"answerValue": question.slug}])
    assert ordered == [doc_c_id, doc_b_id, doc_a_id]

    # Just for coverage - sort by attribute
    ordered = get_ordered([{"attribute": "FORM"}])

    assert ordered == [
        str(doc.id)
        for doc in sorted([doc_c, doc_b, doc_a], key=lambda doc: doc.form.slug)
    ]


@pytest.mark.parametrize("direction", ["ASC", "DESC"])
@pytest.mark.parametrize("question__type", [Question.TYPE_TEXT])
def test_order_by_case_document_answer(
    db, schema_executor, work_item_factory, direction, question
):

    w1, w2 = work_item_factory.create_batch(2)

    w1.case.document.answers.create(question=question, value="aa")
    w2.case.document.answers.create(question=question, value="bb")

    assert w1.case.id != w2.case.id
    assert w1.case.document.id != w2.case.document.id

    query = """
        query workitems {
          allWorkItems(order: [{caseDocumentAnswer: "%s", direction: %s}]) {
            edges {
              node {
                id
              }
            }
          }
        }
    """ % (
        question.slug,
        direction,
    )

    result = schema_executor(query)

    assert not result.errors
    assert result.data["allWorkItems"]["edges"] is not None
    assert len(result.data["allWorkItems"]["edges"]) == 2

    work_items = [
        extract_global_id(wi["node"]["id"])
        for wi in result.data["allWorkItems"]["edges"]
    ]

    if direction == "ASC":
        assert work_items == [str(w1.id), str(w2.id)]
    else:
        assert work_items == [str(w2.id), str(w1.id)]



import pytest

from ...caluma_core.relay import extract_global_id
from ...caluma_form.models import Question
from .. import models


@pytest.mark.parametrize(
    "case__status,result_count",
    [(models.Case.STATUS_RUNNING, 1), (models.Case.STATUS_COMPLETED, 0)],
)
def test_query_all_cases(db, snapshot, case, result_count, flow, schema_executor):
    query = """
        query AllCases {
          allCases (status: RUNNING){
            totalCount
            edges {
              node {
                status
              }
            }
          }
        }
    """
    result = schema_executor(query)

    assert len(result.data["allCases"]["edges"]) == result_count

    assert not result.errors
    snapshot.assert_match(result.data)


@pytest.mark.parametrize("task__lead_time", [100, None])
@pytest.mark.parametrize("task__address_groups", ['["group-name"]|groups', None])
@pytest.mark.parametrize("mutation", ["startCase", "saveCase"])
def test_start_case(
    db,
    snapshot,
    workflow,
    workflow_allow_forms,
    workflow_start_tasks,
    work_item,
    form,
    schema_executor,
    mutation,
    case,
):

    input_type = {"startCase": "StartCaseInput!", "saveCase": "SaveCaseInput!"}[
        mutation
    ]
    query = """
        mutation StartCase($input: %s) {
          %s (input: $input) {
            case {
              id
              document {
                form {
                  slug
                }
              }
              status
              parentWorkItem {
                status
              }
              workItems {
                edges {
                  node {
                    status
                    addressedGroups
                    document {
                      form {
                        slug
                      }
                    }
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """ % (
        input_type,
        mutation,
    )

    inp = {"input": {"workflow": workflow.slug, "form": form.slug}}

    if mutation == "saveCase":
        inp["input"]["id"] = str(case.id)

    result = schema_executor(query, variables=inp)

    assert not result.errors

    # if StartCase, we need a new ID, if "saveCase", we expect the same ID
    id_needs_to_be_same = mutation == "saveCase"
    is_same_id = extract_global_id(result.data[mutation]["case"]["id"]) == str(case.id)
    assert is_same_id == id_needs_to_be_same

    # can't snapshot IDs, they change
    del result.data[mutation]["case"]["id"]
    snapshot.assert_match(result.data)


def test_start_sub_case(db, workflow, work_item, schema_executor):
    query = """
        mutation StartCase($input: StartCaseInput!) {
          startCase(input: $input) {
            case {
              id
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"workflow": workflow.slug, "parentWorkItem": str(work_item.pk)}}
    result = schema_executor(query, variables=inp)

    assert not result.errors

    case_id = result.data["startCase"]["case"]["id"]
    case = models.Case.objects.get(pk=extract_global_id(case_id))
    assert case.parent_work_item.pk == work_item.pk


def test_start_case_invalid_form(db, workflow, form, schema_executor):
    query = """
        mutation StartCase($input: StartCaseInput!) {
          startCase(input: $input) {
            case {
              id
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"workflow": workflow.slug, "form": str(form.pk)}}
    result = schema_executor(query, variables=inp)

    assert result.errors


@pytest.mark.parametrize(
    "work_item__status",
    [models.WorkItem.STATUS_COMPLETED, models.WorkItem.STATUS_READY],
)
@pytest.mark.parametrize(
    "case__status,success",
    [(models.Case.STATUS_RUNNING, True), (models.Case.STATUS_COMPLETED, False)],
)
def test_cancel_case(db, snapshot, case, work_item, schema_executor, success):
    query = """
        mutation CancelCase($input: CancelCaseInput!) {
          cancelCase(input: $input) {
            case {
              document {
                form {
                  slug
                }
              }
              status
              workItems {
                edges {
                  node {
                    status
                  }
                }
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": {"id": case.pk}}
    result = schema_executor(query, variables=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)


@pytest.mark.parametrize(
    "task__is_multiple_instance,task__address_groups,count",
    [(False, ["group1", "group2"], 1), (True, ["group1", "group2"], 2)],
)
def test_multiple_instance_task_address_groups(
    db, workflow, workflow_start_tasks, task, count, schema_executor
):
    query = """
        mutation StartCase($input: StartCaseInput!) {
          startCase(input: $input) {
            clientMutationId
          }
        }
    """

    inp = {"input": {"workflow": workflow.slug}}
    result = schema_executor(query, variables=inp)
    assert not bool(result.errors)
    assert models.WorkItem.objects.count() == count


def test_status_filter(db, case_factory, schema_executor):
    case_factory(status=models.Case.STATUS_CANCELED)
    case_factory(status=models.Case.STATUS_COMPLETED)
    case_factory(status=models.Case.STATUS_RUNNING)

    query = """
        query AllCases {
          allCases (status: [RUNNING, CANCELED]){
            totalCount
            edges {
              node {
                status
              }
            }
          }
        }
    """
    result = schema_executor(query)

    expected = ["CANCELED", "RUNNING"]
    received = sorted(
        [edge["node"]["status"] for edge in result.data["allCases"]["edges"]]
    )

    assert expected == received


@pytest.mark.parametrize("asc", [True, False])
@pytest.mark.parametrize(
    "type,success",
    [
        (Question.TYPE_TEXT, True),
        (Question.TYPE_FORM, True),
        (Question.TYPE_DATE, True),
        (Question.TYPE_FILE, True),
        (Question.TYPE_TABLE, False),
    ],
)
def test_order_by_question_answer_value(
    db,
    snapshot,
    schema_executor,
    asc,
    type,
    success,
    case_factory,
    document_factory,
    question_factory,
    form_question_factory,
    form_factory,
    answer_factory,
):
    value = "test_question1"

    if type == Question.TYPE_TEXT:
        d1 = document_factory()
        d2 = document_factory()
        d3 = document_factory()

        q1 = question_factory(type=Question.TYPE_TEXT, slug="test_question1")
        answer_factory(question=q1, value="c", document=d1)
        answer_factory(question=q1, value="a", document=d2)
        answer_factory(question=q1, value="b", document=d3)

        q2 = question_factory(type=Question.TYPE_TEXT, slug="test_question2")
        answer_factory(question=q2, value="a2", document=d1)
        answer_factory(question=q2, value="b2", document=d2)
        answer_factory(question=q2, value="c2", document=d3)

        case_factory(document=d1)
        case_factory(document=d2)
        case_factory(document=d3)

    elif type == Question.TYPE_FORM:
        value = "test_sub_question1"

        f = form_factory(slug="test_sub_form")
        question = question_factory(type=Question.TYPE_TEXT, slug="test_sub_question1")
        form_question_factory(form=f, question=question)

        d1 = document_factory(form=f)
        answer_factory(question=question, value="d", document=d1)

        d2 = document_factory(form=f)
        answer_factory(question=question, value="b", document=d2)

        d3 = document_factory(form=f)
        answer_factory(question=question, value="f", document=d3)

        form_question = question_factory(
            type=Question.TYPE_FORM, slug="test_form_question", sub_form=f
        )

        form_question_factory(question=form_question)

        case_factory(document=d1)

        case_factory(document=d2)

        case_factory(document=d3)

        # add another form_question and corresponding answers
        form_question_factory(question=question)
        answer_factory(question=question, value="c")
        answer_factory(question=question, value="e")
        answer_factory(question=question, value="a")

    elif type == Question.TYPE_DATE:
        d1 = document_factory()
        d2 = document_factory()
        d3 = document_factory()

        q1 = question_factory(type=Question.TYPE_DATE, slug="test_question1")
        answer_factory(question=q1, date="2019-05-31", document=d1)
        answer_factory(question=q1, date="2019-05-29", document=d2)
        answer_factory(question=q1, date="2019-05-27", document=d3)

        q2 = question_factory(type=Question.TYPE_DATE, slug="test_question2")
        answer_factory(question=q2, date="2019-05-26", document=d1)
        answer_factory(question=q2, date="2019-05-30", document=d2)
        answer_factory(question=q2, date="2019-05-28", document=d3)

        case_factory(document=d1)
        case_factory(document=d2)
        case_factory(document=d3)

    elif type == Question.TYPE_FILE:
        d1 = document_factory()
        d2 = document_factory()
        d3 = document_factory()

        q1 = question_factory(type=Question.TYPE_FILE, slug="test_question1")
        answer_factory(question=q1, file__name="d", document=d1)
        answer_factory(question=q1, file__name="b", document=d2)
        answer_factory(question=q1, file__name="f", document=d3)

        q2 = question_factory(type=Question.TYPE_FILE, slug="test_question2")
        answer_factory(question=q2, file__name="c", document=d1)
        answer_factory(question=q2, file__name="e", document=d2)
        answer_factory(question=q2, file__name="a", document=d3)

        case_factory(document=d1)
        case_factory(document=d2)
        case_factory(document=d3)

    elif type == Question.TYPE_TABLE:
        d1 = document_factory()
        question_factory(type=Question.TYPE_TABLE, slug="test_question1")
        case_factory(document=d1)

    # It's necessary to order the answers by "CREATED_AT_ASC" in order to every time
    # produce the same response
    query = """
        query AllCases($orderByQuestionAnswerValue: String) {
          allCases(orderByQuestionAnswerValue: $orderByQuestionAnswerValue){
            totalCount
            edges {
              node {
                document {
                  answers(orderBy: CREATED_AT_ASC) {
                    totalCount
                    edges {
                      node {
                        ... on StringAnswer {
                          stringValue: value
                        }
                        ... on DateAnswer {
                          dateValue: value
                        }
                        ... on FileAnswer {
                          fileValue: value {
                            name
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
    """

    if not asc:
        value = f"-{value}"

    inp = {"orderByQuestionAnswerValue": value}

    result = schema_executor(query, variables=inp)

    assert not bool(result.errors) == success
    if success:
        assert result.data["allCases"]["totalCount"] == 3
        snapshot.assert_match(result.data)


def test_document_form(
    db, schema_executor, case_factory, document_factory, form_factory
):

    form_a, form_b = form_factory.create_batch(2)

    # two cases, one for each form
    case_factory(document=document_factory(form=form_a))
    case_factory(document=document_factory(form=form_b))

    query = """
        query AllCases ($form: String!) {
          allCases (documentForm: $form){
            totalCount
            edges {
              node {
                document {
                  form {
                    id
                  }
                }
              }
            }
          }
        }
    """
    # search for form A's slug
    result = schema_executor(query, variables={"form": form_a.slug})

    assert len(result.data["allCases"]["edges"]) == 1

    assert not result.errors

    document = result.data["allCases"]["edges"][0]["node"]["document"]

    assert extract_global_id(document["form"]["id"]) == form_a.slug


@pytest.mark.parametrize("question__slug,question__type", [("asdf", "text")])
@pytest.mark.parametrize("form__slug", ["theform"])
@pytest.mark.parametrize(
    "doc_val,workitem_val,result_count", [("hello", "blah", 0), ("blah", "hello", 1)]
)
def test_work_item_document(
    db,
    case,
    form,
    work_item_factory,
    answer_factory,
    question,
    schema_executor,
    doc_val,
    workitem_val,
    result_count,
):
    work_item = work_item_factory(document__form=form, case=case)
    answer_factory(document=work_item.document, question=question, value=workitem_val)
    answer_factory(document=case.document, question=question, value=doc_val)

    query = """
        query workitems ($filter: HasAnswerFilterType){
          allCases(filter: [
            {workItemDocumentHasAnswer: [$filter]}
          ]) {
            edges {
              node {
                id
              }
            }
          }
        }

    """
    result = schema_executor(
        query,
        variables={
            "filter": {"question": question.slug, "value": "hello", "lookup": "EXACT"}
        },
    )

    assert not result.errors

    assert len(result.data["allCases"]["edges"]) == result_count



import pytest
from rest_framework.exceptions import ValidationError

from ...caluma_core.tests import extract_serializer_input_fields
from ...caluma_form.models import DynamicOption, Question
from .. import serializers, structure
from ..jexl import QuestionMissing
from ..validators import DocumentValidator, QuestionValidator


@pytest.mark.parametrize(
    "required_jexl,hidden_jexl,should_throw",
    [
        ("false", "true", False),
        ("false", "false", False),
        ("true", "true", False),
        ("true", "false", True),
    ],
)
@pytest.mark.parametrize("question__type", [Question.TYPE_TEXT])
def test_validate_hidden_required_field(
    db, required_jexl, hidden_jexl, should_throw, form_question, document_factory, info
):
    form_question.question.is_required = required_jexl
    form_question.question.is_hidden = hidden_jexl
    form_question.question.save()

    document = document_factory(form=form_question.form)
    error_msg = f"Questions {form_question.question.slug} are required but not provided"
    if should_throw:
        with pytest.raises(ValidationError, match=error_msg):
            DocumentValidator().validate(document, info)
    else:
        DocumentValidator().validate(document, info)


@pytest.mark.parametrize(
    "question__type,question__is_required",
    [(Question.TYPE_FILE, "false"), (Question.TYPE_DATE, "false")],
)
def test_validate_special_fields(
    db, form_question, question, document_factory, answer_factory, info
):
    document = document_factory(form=form_question.form)
    answer_factory(document=document, question=question)
    DocumentValidator().validate(document, info)


@pytest.mark.parametrize(
    "question__data_source,value,valid",
    [
        ("MyDataSource", "5.5", True),
        ("MyDataSource", 5.5, False),
        ("MyOtherDataSource", "5.5", True),
        ("MyOtherDataSource", 5.5, False),
        ("MyOtherDataSource", 23, False),
        ("MyDataSource", "not in data", False),
        ("MyOtherDataSource", "not in data", True),
    ],
)
@pytest.mark.parametrize(
    "question__type",
    [Question.TYPE_DYNAMIC_CHOICE, Question.TYPE_DYNAMIC_MULTIPLE_CHOICE],
)
def test_validate_dynamic_options(
    db,
    form_question,
    question,
    value,
    valid,
    document_factory,
    answer_factory,
    info,
    settings,
):
    settings.DATA_SOURCE_CLASSES = [
        "caluma.caluma_data_source.tests.data_sources.MyDataSource",
        "caluma.caluma_data_source.tests.data_sources.MyOtherDataSource",
    ]
    lookup_value = value
    if question.type == Question.TYPE_DYNAMIC_MULTIPLE_CHOICE and not value == 23:
        value = [value]

    document = document_factory(form=form_question.form)
    answer_factory(value=value, document=document, question=question)

    if valid:
        DocumentValidator().validate(document, info)
        assert DynamicOption.objects.get(
            document=document, slug=lookup_value, question=question
        )
    else:
        with pytest.raises(ValidationError):
            DocumentValidator().validate(document, info)


@pytest.mark.parametrize(
    "question__type",
    [Question.TYPE_DYNAMIC_CHOICE, Question.TYPE_DYNAMIC_MULTIPLE_CHOICE],
)
@pytest.mark.parametrize("question__data_source", ["MyDataSource"])
def test_validate_dynamic_option_exists(
    db,
    form_question,
    question,
    answer_factory,
    document_factory,
    dynamic_option_factory,
    info,
    settings,
):
    settings.DATA_SOURCE_CLASSES = [
        "caluma.caluma_data_source.tests.data_sources.MyDataSource"
    ]

    value = "foobar"
    document = document_factory(form=form_question.form)
    dynamic_option = dynamic_option_factory(
        document=document, question=question, slug=value, label="test"
    )

    if question.type == Question.TYPE_DYNAMIC_MULTIPLE_CHOICE:
        value = [value]
    answer_factory(question=question, value=value, document=document)

    assert DocumentValidator().validate(dynamic_option.document, info) is None


@pytest.mark.parametrize(
    "required_jexl_main,required_jexl_sub,should_throw",
    [
        ("true", "false", True),
        ("'other_q_1'|answer == 'something'", "false", True),
        ("'other_q_1'|answer == 'something else'", "false", False),
        ("false", "false", False),
        ("'foo' in 'main_table_1'|answer|mapby('sub_question_a')", "false", True),
        ("'nothere' in 'main_table_1'|answer|mapby('sub_question_a')", "false", False),
        ("false", "'foo' == 'sub_question_a'|answer", True),
        ("false", "'bar' == 'sub_question_a'|answer", False),
        ("false", "'other_q_1'|answer == 'something'", True),
        ("false", "'other_q_1'|answer == 'something else'", False),
        ("false", "'fail' == 'no-question-slug'|answer", True),
    ],
)
def test_validate_table(
    db,
    required_jexl_main,
    required_jexl_sub,
    should_throw,
    answer_document_factory,
    form_factory,
    form_question_factory,
    document_factory,
    answer_factory,
    info,
):

    # F: main-form
    #     Q: main_table_1: table
    #        ROW_FORM
    #           Q: sub_question_a
    #           Q: sub_question_b (required_jexl_sub)
    #     Q: other_q_1
    #     Q: other_q_2 (required_jexl_main)

    main_form = form_factory(slug="main-form")
    main_table_question_1 = form_question_factory(
        form=main_form,
        question__type=Question.TYPE_TABLE,
        question__slug="main_table_1",
        question__is_required="true",
    )

    main_table_question_1.question.row_form = form_factory()
    main_table_question_1.question.save()

    sub_question_a = form_question_factory(
        form=main_table_question_1.question.row_form,
        question__type=Question.TYPE_TEXT,
        question__slug="sub_question_a",
    )
    sub_question_b = form_question_factory(
        form=main_table_question_1.question.row_form,
        question__type=Question.TYPE_TEXT,
        question__is_required=required_jexl_sub,
        question__slug="sub_question_b",
    )
    other_q_1 = form_question_factory(
        form=main_form,
        question__type=Question.TYPE_TEXT,
        question__slug="other_q_1",
        question__is_required="false",
    )

    other_q_2 = form_question_factory(
        form=main_form,
        question__type=Question.TYPE_TEXT,
        question__slug="other_q_2",
        question__is_required=required_jexl_main,
    )

    # MD
    #   - TA
    #       - RD1
    #           sqa:"foo"
    #   oq1:"something"

    main_document = document_factory(form=main_form)
    table_answer = answer_factory(
        document=main_document, question=main_table_question_1.question, value=None
    )

    row_document_1 = document_factory(form=main_table_question_1.question.row_form)
    answer_document_factory(document=row_document_1, answer=table_answer)

    answer_factory(
        question=sub_question_a.question, document=row_document_1, value="foo"
    )
    answer_factory(
        question=other_q_1.question, document=main_document, value="something"
    )

    if should_throw and required_jexl_sub.startswith("'fail'"):
        with pytest.raises(QuestionMissing):
            DocumentValidator().validate(main_document, info)
    elif should_throw:
        q_slug = sub_question_b.question.slug
        if required_jexl_sub == "false":
            q_slug = other_q_2.question.slug
        error_msg = f"Questions {q_slug} are required but not provided"
        with pytest.raises(ValidationError, match=error_msg):
            DocumentValidator().validate(main_document, info)
    else:
        DocumentValidator().validate(main_document, info)


@pytest.mark.parametrize(
    "question__data_source,valid", [("MyDataSource", True), ("NotADataSource", False)]
)
@pytest.mark.parametrize(
    "question__type,serializer_to_use",
    [
        (Question.TYPE_DYNAMIC_CHOICE, "SaveDynamicChoiceQuestionSerializer"),
        (
            Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            "SaveDynamicMultipleChoiceQuestionSerializer",
        ),
    ],
)
def test_validate_data_source(
    db, question, valid, info, serializer_to_use, data_source_settings
):
    serializer = getattr(serializers, serializer_to_use)

    data = extract_serializer_input_fields(serializer, question)
    data["type"] = question.type

    if valid:
        QuestionValidator().validate(data)
    else:
        with pytest.raises(ValidationError):
            QuestionValidator().validate(data)


@pytest.mark.parametrize(
    "question__type,answer__value,expected_value",
    [
        (Question.TYPE_MULTIPLE_CHOICE, None, []),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, None, []),
        (Question.TYPE_TEXT, None, None),
        (Question.TYPE_TEXT, "", ""),
        (Question.TYPE_INTEGER, None, None),
        (Question.TYPE_FLOAT, None, None),
        (Question.TYPE_DATE, None, None),
        (Question.TYPE_FILE, None, None),
        (Question.TYPE_TEXTAREA, None, None),
    ],
)
@pytest.mark.parametrize("answer__date,answer__file", [(None, None)])
def test_validate_empty_answers(
    db,
    form_question,
    document,
    answer,
    question,
    document_factory,
    answer_factory,
    expected_value,
    info,
):

    struct = structure.FieldSet(document, document.form)
    field = struct.get_field(question.slug)
    answer_value = field.value() if field else field
    assert answer_value == expected_value


@pytest.mark.parametrize(
    "question__slug,question__type,question__is_required,question__is_hidden,answer__value,exception_message",
    [
        (
            "q-slug",
            Question.TYPE_TEXT,
            "'foo' in blah",
            "false",
            "",
            "Error while evaluating `is_required` expression on question q-slug: 'foo' in blah. The system log contains more information",
        ),
        (
            "q-slug",
            Question.TYPE_TEXT,
            "true",
            "'foo' in blah",
            "",
            "Error while evaluating `is_hidden` expression on question q-slug: 'foo' in blah. The system log contains more information",
        ),
        (
            "q-slug",
            Question.TYPE_TEXT,
            "'value' == 'q-slug'|answer",
            "false",
            "value",
            None,
        ),
    ],
)
def test_validate_invalid_jexl(
    db, form_question, document, answer, question, exception_message, info
):

    if exception_message is not None:
        with pytest.raises(RuntimeError) as exc:
            DocumentValidator().validate(document, info)
        assert exc.value.args[0] == exception_message
    else:
        assert DocumentValidator().validate(document, info) is None


def test_validate_required_integer_0(
    db, form_question, answer_factory, document_factory, info
):
    form_question.question.is_required = "true"
    form_question.question.type = Question.TYPE_INTEGER
    form_question.question.save()

    document = document_factory(form=form_question.form)
    answer_factory(document=document, value=0, question=form_question.question)

    DocumentValidator().validate(document, info)


@pytest.mark.parametrize(
    "question__type,answer__value",
    [
        (Question.TYPE_MULTIPLE_CHOICE, None),
        (Question.TYPE_MULTIPLE_CHOICE, []),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, None),
        (Question.TYPE_DYNAMIC_MULTIPLE_CHOICE, []),
        (Question.TYPE_TEXT, None),
        (Question.TYPE_TEXT, ""),
        (Question.TYPE_INTEGER, None),
        (Question.TYPE_FLOAT, None),
        (Question.TYPE_DATE, None),
        (Question.TYPE_FILE, None),
        (Question.TYPE_TEXTAREA, None),
        (Question.TYPE_TABLE, None),
        (Question.TYPE_TABLE, []),
        (Question.TYPE_CHOICE, None),
        (Question.TYPE_DYNAMIC_CHOICE, None),
    ],
)
@pytest.mark.parametrize(
    "answer__date,answer__file,question__is_required", [(None, None, "true")]
)
def test_validate_required_empty_answers(
    db, info, form_question, document, answer, question
):
    with pytest.raises(ValidationError):
        DocumentValidator().validate(document, info)


@pytest.mark.parametrize("question__is_hidden", ["true", "false"])
@pytest.mark.parametrize(
    "question__type,question__configuration",
    [[Question.TYPE_FLOAT, {"min_value": 0, "max_value": 3}]],
)
def test_validate_hidden_field(
    db, form_question, document_factory, answer_factory, info
):
    question = form_question.question
    document = document_factory(form=form_question.form)

    # answer is out of validation range so should fail
    answer_factory(question=question, document=document, value=4)

    if question.is_hidden == "true":
        assert DocumentValidator().validate(document, info) is None
    else:
        with pytest.raises(ValidationError):
            DocumentValidator().validate(document, info)


@pytest.mark.parametrize("hide_formquestion", [True, False])
def test_validate_hidden_subform(
    db,
    question_factory,
    form_factory,
    document_factory,
    form_question_factory,
    answer_factory,
    info,
    hide_formquestion,
):
    # First, build our nested form:
    #     top_form
    #       \__ form_question # hidden or not/parametrized
    #            \__ sub_form
    #                 \__ sub_question
    #                 \__ sub_sub_fq  # always hidden. sub_sub_question should never be checked
    #                      \__ sub_sub_form
    #                          \__ sub_sub_question # required = true
    top_form = form_factory()
    sub_form = form_factory()
    form_question = question_factory(
        type=Question.TYPE_FORM,
        sub_form=sub_form,
        is_hidden=str(hide_formquestion).lower(),
        slug="form_question",
    )
    top_form.questions.add(form_question)
    sub_question = question_factory(
        type=Question.TYPE_FLOAT,
        configuration={"min_value": 0, "max_value": 3},
        slug="sub_question",
    )
    form_question_factory(form=sub_form, question=sub_question)

    sub_sub_form = form_factory(slug="sub_sub_form")
    sub_sub_fq = question_factory(
        type=Question.TYPE_FORM,
        sub_form=sub_sub_form,
        is_hidden="true",
        slug="sub_sub_fq",
    )
    form_question_factory(form=sub_form, question=sub_sub_fq)

    sub_sub_question = question_factory(
        is_required="true",
        is_hidden="false",
        slug="sub_sub_question",
        type=Question.TYPE_FLOAT,
        configuration={"min_value": 1, "max_value": 3},
    )
    form_question_factory(form=sub_sub_form, question=sub_sub_question)

    # Second, make a document. The answer for the sub_question should
    # be invalid to test if the hidden form question masks it properly.
    # We never answer sub_sub_question, thus making the document invalid
    # if it's wrongfully validated
    document = document_factory(form=top_form)
    answer_factory(question=sub_sub_question, document=document, value=4)

    if hide_formquestion:
        assert DocumentValidator().validate(document, info) is None
    else:
        with pytest.raises(ValidationError) as excinfo:
            DocumentValidator().validate(document, info)
        # Verify that the sub_sub_question is not the cause of the exception:
        # it should not be checked at all because it's parent is always hidden
        assert excinfo.match(r"Questions \bsub_question\s.* required but not provided.")

        # can't do `not excinfo.match()` as it throws an AssertionError itself
        # if it can't match :()
        assert "sub_sub_question" not in str(excinfo.value.detail)


@pytest.mark.parametrize("answer_value", ["foo", "bar"])
def test_dependent_question_is_hidden(
    db,
    question_factory,
    form_factory,
    document_factory,
    form_question_factory,
    answer_factory,
    answer_value,
    info,
):
    form = form_factory()
    q1 = question_factory(is_hidden="false", type=Question.TYPE_TEXT)
    q2 = question_factory(
        is_hidden=f"'{q1.slug}'|answer=='foo'",
        type=Question.TYPE_TEXT,
        is_required="true",
    )
    q3 = question_factory(
        is_hidden=f"'{q2.slug}'|answer=='bar'",
        is_required="true",
        type=Question.TYPE_TEXT,
    )
    form_question_factory(form=form, question=q1)
    form_question_factory(form=form, question=q2)
    form_question_factory(form=form, question=q3)

    document = document_factory(form=form)
    answer_factory(question=q1, document=document, value=answer_value)
    answer_factory(question=q2, document=document, value="foo")

    if answer_value == "foo":
        # Answer to q1's value is "foo", so q2 is hidden. This means
        # that q3 should be hidden and not required as well
        assert DocumentValidator().validate(document, info) is None
    else:
        # a1's value is "bar", so q2 is visible. This means
        # that q3 should also be visible and required
        with pytest.raises(ValidationError):
            DocumentValidator().validate(document, info)


@pytest.mark.parametrize("question__type", ["file"])
@pytest.mark.parametrize("question__is_required", ["true"])
@pytest.mark.parametrize("question__is_hidden", ["false"])
def test_required_file(db, question, form, document, answer, info, form_question):
    # verify some assumptions
    assert document.form == form
    assert answer.document == document
    assert form in question.forms.all()
    assert answer.file

    # remove the file
    the_file = answer.file
    answer.file = None
    answer.save()

    # ensure validation fails with no `file` value
    with pytest.raises(ValidationError):
        DocumentValidator().validate(document, info)

    # put the file back
    answer.file = the_file
    answer.save()

    # then, validation must pass
    assert DocumentValidator().validate(document, info) is None


def test_validate_missing_in_subform(
    db,
    question_factory,
    form_factory,
    document_factory,
    form_question_factory,
    answer_factory,
    info,
):
    # First, build our nested form:
    #     top_form
    #       \__ form_question # not required
    #            \__ sub_form
    #                 \__ sub_question# required, but not filled
    top_form = form_factory()
    sub_form = form_factory()

    # form question is not required
    form_question = question_factory(
        type=Question.TYPE_FORM,
        sub_form=sub_form,
        is_hidden="false",
        is_required="false",
        slug="form_question",
    )
    form_question_factory(form=top_form, question=form_question)

    # sub question is required
    sub_question = question_factory(
        type=Question.TYPE_FLOAT,
        configuration={"min_value": 0, "max_value": 3},
        is_hidden="false",
        is_required="true",
        slug="sub_question",
    )
    form_question_factory(form=sub_form, question=sub_question)

    # Second, make a document.
    # the sub question remains unanswered, which should raise an error
    document = document_factory(form=top_form)

    with pytest.raises(ValidationError) as excinfo:
        DocumentValidator().validate(document, info)

    # Verify that the sub_sub_question is not the cause of the exception:
    # it should not be checked at all because it's parent is always hidden
    assert excinfo.match(r"Questions \bsub_question\s.* required but not provided.")


@pytest.mark.parametrize("table_required", [True, False])
def test_validate_missing_in_table(
    db,
    question_factory,
    form_factory,
    document_factory,
    form_question_factory,
    answer_document_factory,
    answer_factory,
    table_required,
    info,
):
    # First, build our nested form:
    #     top_form
    #       \__ table_question # requiredness parametrized
    #            \__ sub_form
    #                 \__ sub_question1 # required and filled
    #                 \__ sub_question2 # required, but not filled
    top_form = form_factory()
    sub_form = form_factory()

    # form question is not required
    table_question = question_factory(
        type=Question.TYPE_TABLE,
        row_form=sub_form,
        is_hidden="false",
        is_required=str(table_required).lower(),
        slug="table_question",
    )
    form_question_factory(form=top_form, question=table_question)

    # sub question1 is NOT required
    sub_question1 = question_factory(
        type=Question.TYPE_TEXT,
        is_hidden="false",
        is_required="true",
        slug="sub_question1",
    )
    # sub question1 is required
    sub_question2 = question_factory(
        type=Question.TYPE_TEXT,
        is_hidden="false",
        is_required="true",
        slug="sub_question2",
    )
    form_question_factory(form=sub_form, question=sub_question1)
    form_question_factory(form=sub_form, question=sub_question2)

    # Second, make a document.
    # it has one row, but the answer to question2 is missing
    document = document_factory(form=top_form)
    table_ans = answer_factory(document=document, question=table_question)

    if table_required:
        # Table required. But we only fill it partially, this
        # should raise an error
        row_doc = document_factory(form=sub_form)
        answer_document_factory(answer=table_ans, document=row_doc)
        row_doc.answers.create(question=sub_question1, value="hi")

        with pytest.raises(ValidationError) as excinfo:
            DocumentValidator().validate(document, info)
        # Verify that the sub_sub_question is not the cause of the exception:
        # it should not be checked at all because it's parent is always hidden
        assert excinfo.match(
            r"Questions \bsub_question2\s.* required but not provided."
        )

    else:
        # should not raise
        DocumentValidator().validate(document, info)



from dateparser import parse
from django.core.management.base import BaseCommand
from django.utils import timezone
from simple_history.models import registered_models


class Command(BaseCommand):
    """Cleanup historical records."""

    help = "Cleanup historical records."

    def add_arguments(self, parser):
        parser.add_argument("--force", dest="force", default=False, action="store_true")
        parser.add_argument(
            "-k",
            "--keep",
            dest="keep",
            default="1 year",
            help=(
                "Duration we want to keep the records. "
                "E.g. '6 months', '1 year'. Uses dateparser."
            ),
        )

    def handle(self, *args, **options):
        force = options["force"]

        lt = parse(options["keep"])
        if lt is not None:
            lt = timezone.make_aware(lt)

        for _, model in registered_models.items():
            qs = model.history.all()
            if lt is not None:
                qs = model.history.filter(history_date__lt=lt)

            action_str = "Deleting" if force else "Would delete"
            self.stdout.write(
                f'{action_str} {qs.count()} historical records from model "{model.__name__}"'
            )
            if force:
                qs.delete()



# Generated by Django 2.2.6 on 2019-10-31 12:22

from django.db import migrations

import caluma.caluma_core.models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0014_add_gin_index_to_jsonfields")]

    operations = [
        migrations.AlterField(
            model_name="historicalworkitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Task is ready to be processed."),
                    ("completed", "Task is done."),
                    ("canceled", "Task is cancelled."),
                    ("skipped", "Task is skipped."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="workitem",
            name="status",
            field=caluma.caluma_core.models.ChoicesCharField(
                choices=[
                    ("ready", "Task is ready to be processed."),
                    ("completed", "Task is done."),
                    ("canceled", "Task is cancelled."),
                    ("skipped", "Task is skipped."),
                ],
                db_index=True,
                max_length=50,
            ),
        ),
    ]



from collections import defaultdict
from functools import partial

from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context

from ..caluma_core.jexl import JEXL, ExtractTransformSubjectAnalyzer


class QuestionMissing(Exception):
    pass


class QuestionValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "answer" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid question slug."

        yield from super().visit_Transform(transform)


class QuestionJexl(JEXL):
    def __init__(self, validation_context=None, **kwargs):

        if validation_context:
            if "jexl_cache" not in validation_context:
                validation_context["jexl_cache"] = defaultdict(dict)
            self._cache = validation_context["jexl_cache"]
        else:
            self._cache = defaultdict(dict)

        super().__init__(**kwargs)

        context_data = None
        if validation_context:
            context_data = {**validation_context}
            form = validation_context.get("form")
            if form:
                context_data["form"] = form.slug

        self.context = Context(context_data)

        self.add_transform("answer", self.answer_transform)
        self.add_transform("mapby", lambda arr, key: [obj[key] for obj in arr])
        self.add_binary_operator(
            "intersects", 20, lambda left, right: any(x in right for x in left)
        )

    def answer_transform(self, question):
        return self.context["structure"].get_field(question).value()

    def validate(self, expression, **kwargs):
        return super().validate(expression, QuestionValidatingAnalyzer)

    def extract_referenced_questions(self, expr):
        transforms = ["answer", "mapby"]
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=transforms)
        )

    def _question(self, slug):
        field = self.context["structure"].get_field(slug)
        if field:
            return field.question

        raise QuestionMissing(
            f"Question `{slug}` could not be found in form {self.context['form']}"
        )

    def _all_containers_hidden(self, question):
        """Check whether all containers of the given question are hidden.

        The question could be used in multiple sub-forms in the given
        document structure. This goes through all paths from the main form
        to the given question, and checks if they're hidden.

        If all subforms are hidden where the question shows up,
        the question shall be hidden as well.
        """
        paths = self.context.get("structure").paths_to_question(question.slug)

        res = bool(paths) and all(
            # the "inner" check here represents a single path. If any
            # question is hidden, the question is not reachable via
            # this path
            bool(path) and any(self.is_hidden(fq) for fq in path if fq != question)
            for path in paths
        )
        # The outer check verifies if all paths are "hidden" (ie have a hidden
        # question in them). If all paths from the root form to the
        # question are hidden, the question must indeed be hidden itself.

        return res

    def is_hidden(self, question):
        """Return True if the given question is hidden.

        This checks whether the dependency questions are hidden, then
        evaluates the question's is_hidden expression itself.
        """

        if question.pk in self._cache["hidden"]:
            return self._cache["hidden"][question.pk]
        # Check visibility of dependencies before actually evaluating the `is_hidden`
        # expression. If all dependencies are hidden,
        # there is no way to evaluate our own visibility, so we default to
        # hidden state as well.

        deps = list(self.extract_referenced_questions(question.is_hidden))

        # all() returns True for the empty set, thus we need to
        # check that we have some deps at all first
        all_deps_hidden = bool(deps) and all(
            self.is_hidden(self._question(dep)) for dep in deps
        )
        if all_deps_hidden:
            self._cache["hidden"][question.pk] = True
            return True

        # Also check if the question is hidden indirectly,
        # for example via parent formquestion.
        if self._all_containers_hidden(question):
            # no way this is shown somewhere
            self._cache["hidden"][question.pk] = True
            return True
        # if the question is visible-in-context and not hidden by invisible dependencies,
        # we can evaluate it's own is_hidden expression
        self._cache["hidden"][question.pk] = self.evaluate(question.is_hidden)
        return self._cache["hidden"][question.pk]

    def is_required(self, question):
        if question.pk in self._cache["required"]:
            return self._cache["required"][question.pk]

        deps = list(self.extract_referenced_questions(question.is_required))

        if bool(deps) and all(self.is_hidden(self._question(dep)) for dep in deps):
            # all dependent questions are hidden. cannot evaluate,
            # so assume requiredness to be False
            ret = False
        else:
            ret = self.evaluate(question.is_required)
        self._cache["required"][question.pk] = ret
        return ret
